import os.path

import requests

cache_root = os.path.join(os.environ['HOME'], 'notecache')


def url_to_path(url: str):
    if url.startswith("https://raw.githubusercontent.com/notechats"):
        base_url = url.split("https://raw.githubusercontent.com/")[1]
        base_path, filename = os.path.split(base_url)
        cache_path = os.path.join(cache_root, base_path)
        file_path = os.path.join(cache_path, filename)

        # 如果缓存路径不存在则创建
        if not os.path.exists(cache_path):
            os.makedirs(cache_path)

        return file_path
    return None


def load_url_to_cache(url: str, overwrite=False) -> str:
    file_path = url_to_path(url)
    if file_path is None:
        return file_path

    # 如果文件已存在
    if os.path.exists(file_path):
        # 如果不覆盖就返回，覆盖就删除原文件
        if not overwrite:
            return file_path
        else:
            os.remove(file_path)

    with open(file_path, 'wb') as f:
        content = requests.get(url).content
        f.write(content)
    return file_path


def load_from_cache(url: str):
    file_path = load_url_to_cache(url)
    return open(file_path).read()



from gevent import monkey

monkey.patch_all()

from gevent.server import DatagramServer
import gevent
from hashlib import sha1
from random import randint
from struct import unpack
from socket import inet_ntoa
from gevent import sleep
from collections import deque
from .bencode import bencode, bdecode

BOOTSTRAP_NODES = (
    ("router.bittorrent.com", 6881),
    ("dht.transmissionbt.com", 6881),
    ("router.utorrent.com", 6881)
)
TID_LENGTH = 2
RE_JOIN_DHT_INTERVAL = 3
MONITOR_INTERVAL = 10
TOKEN_LENGTH = 2


def entropy(length):
    return "".join(chr(randint(0, 255)) for _ in xrange(length))


def random_id():
    h = sha1()
    h.update(entropy(20))
    return h.digest()


def decode_nodes(nodes):
    n = []
    length = len(nodes)
    if (length % 26) != 0:
        return n

    for i in range(0, length, 26):
        nid = nodes[i:i + 20]
        ip = inet_ntoa(nodes[i + 20:i + 24])
        port = unpack("!H", nodes[i + 24:i + 26])[0]
        n.append((nid, ip, port))

    return n


def get_neighbor(target, nid, end=10):
    return target[:end] + nid[end:]


class KNode(object):

    def __init__(self, nid, ip, port):
        self.nid = nid
        self.ip = ip
        self.port = port


class DHTServer(DatagramServer):
    def __init__(self, max_node_qsize, bind_ip):
        s = ':' + str(bind_ip)
        self.bind_ip = bind_ip
        DatagramServer.__init__(self, s)

        self.process_request_actions = {
            "get_peers": self.on_get_peers_request,
            "announce_peer": self.on_announce_peer_request,
        }
        self.max_node_qsize = max_node_qsize
        self.nid = random_id()
        self.nodes = deque(maxlen=max_node_qsize)

    def handle(self, data, address):  #
        try:
            msg = bdecode(data)
            self.on_message(msg, address)
        except Exception:
            pass

    def monitor(self):
        while True:
            # print 'len: ', len(self.nodes)
            sleep(MONITOR_INTERVAL)

    def send_krpc(self, msg, address):
        try:
            self.socket.sendto(bencode(msg), address)
        except Exception:
            pass

    def send_find_node(self, address, nid=None):
        nid = get_neighbor(nid, self.nid) if nid else self.nid
        tid = entropy(TID_LENGTH)
        msg = {
            "t": tid,
            "y": "q",
            "q": "find_node",
            "a": {
                "id": nid,
                "target": random_id()
            }
        }
        self.send_krpc(msg, address)

    def join_DHT(self):
        for address in BOOTSTRAP_NODES:
            self.send_find_node(address)

    def re_join_DHT(self):

        while True:
            if len(self.nodes) == 0:
                self.join_DHT()
            sleep(RE_JOIN_DHT_INTERVAL)

    def auto_send_find_node(self):

        wait = 1.0 / self.max_node_qsize / 5.0
        while True:
            try:
                node = self.nodes.popleft()
                self.send_find_node((node.ip, node.port), node.nid)
            except IndexError:
                pass
            sleep(wait)

    def process_find_node_response(self, msg, address):
        # print 'find node' + str(msg)
        nodes = decode_nodes(msg["r"]["nodes"])
        for node in nodes:
            (nid, ip, port) = node
            if len(nid) != 20: continue
            if ip == self.bind_ip: continue
            if port < 1 or port > 65535: continue
            n = KNode(nid, ip, port)
            self.nodes.append(n)

    def on_message(self, msg, address):
        try:
            if msg["y"] == "r":
                if msg["r"].has_key("nodes"):
                    self.process_find_node_response(msg, address)
            elif msg["y"] == "q":
                try:
                    self.process_request_actions[msg["q"]](msg, address)
                except KeyError:
                    self.play_dead(msg, address)
        except KeyError:
            pass

    def on_get_peers_request(self, msg, address):
        try:
            infohash = msg["a"]["info_hash"]
            tid = msg["t"]
            nid = msg["a"]["id"]
            token = infohash[:TOKEN_LENGTH]
            info = infohash.encode("hex").upper() + '|' + address[0]
            print
            info + "\n",
            msg = {
                "t": tid,
                "y": "r",
                "r": {
                    "id": get_neighbor(infohash, self.nid),
                    "nodes": "",
                    "token": token
                }
            }
            self.send_krpc(msg, address)
        except KeyError:
            pass

    def on_announce_peer_request(self, msg, address):
        try:
            # print 'announce peer'
            infohash = msg["a"]["info_hash"]
            token = msg["a"]["token"]
            nid = msg["a"]["id"]
            tid = msg["t"]

            if infohash[:TOKEN_LENGTH] == token:
                if msg["a"].has_key("implied_port") and msg["a"]["implied_port"] != 0:
                    port = address[1]
                else:
                    port = msg["a"]["port"]
                    if port < 1 or port > 65535: return
                info = infohash.encode("hex").upper()
                print
                info + "\n",
        except Exception as e:
            print
            e
            pass
        finally:
            self.ok(msg, address)

    def play_dead(self, msg, address):
        try:
            tid = msg["t"]
            msg = {
                "t": tid,
                "y": "e",
                "e": [202, "Server Error"]
            }
            self.send_krpc(msg, address)
        except KeyError:
            print
            'error'
            pass

    def ok(self, msg, address):
        try:
            tid = msg["t"]
            nid = msg["a"]["id"]
            msg = {
                "t": tid,
                "y": "r",
                "r": {
                    "id": get_neighbor(nid, self.nid)
                }
            }
            self.send_krpc(msg, address)
        except KeyError:
            pass


if __name__ == '__main__':
    sniffer = DHTServer(50, 8080)
    gevent.spawn(sniffer.auto_send_find_node)
    gevent.spawn(sniffer.re_join_DHT)
    gevent.spawn(sniffer.monitor)

    print('Receiving datagrams on :6882')
    sniffer.serve_forever()



import six


class BTFailure(Exception):
    pass


def decode_int(x, f):
    f += 1
    newf = x.find(b"e", f)
    n = int(x[f:newf])
    if six.indexbytes(x, f) == 45:
        if six.indexbytes(x, f + 1) == 48:
            raise ValueError
    elif six.indexbytes(x, f) == 48 and newf != f + 1:
        raise ValueError
    return n, newf + 1


def decode_string(x, f):
    colon = x.find(b":", f)
    n = int(x[f:colon])
    if six.indexbytes(x, f) == 48 and colon != f + 1:
        raise ValueError
    colon += 1
    return x[colon: colon + n], colon + n


def decode_list(x, f):
    r, f = [], f + 1
    while six.indexbytes(x, f) != 101:
        v, f = decode_func[six.indexbytes(x, f)](x, f)
        r.append(v)
    return r, f + 1


def decode_dict(x, f):
    r, f = {}, f + 1
    while six.indexbytes(x, f) != 101:
        k, f = decode_string(x, f)
        r[k], f = decode_func[six.indexbytes(x, f)](x, f)
    return r, f + 1


decode_func = {
    108: decode_list,
    100: decode_dict,
    105: decode_int
}

for i in range(48, 59):
    decode_func[i] = decode_string


def bdecode(x):
    try:
        r, l = decode_func[six.indexbytes(x, 0)](x, 0)
    except (IndexError, KeyError, ValueError):
        raise BTFailure("not a valid bencoded string")
    if l != len(x):
        raise BTFailure("invalid bencoded value (data after valid prefix)")
    return r


class Bencached(object):
    __slots__ = ["bencoded"]

    def __init__(self, s):
        self.bencoded = s


def encode_bencached(x, r):
    r.append(x.bencoded)


def encode_int(x, r):
    r.extend((b"i", str(x).encode(), b"e"))


def encode_bool(x, r):
    if x:
        encode_int(1, r)
    else:
        encode_int(0, r)


def encode_string(x, r):
    r.extend((str(len(x)).encode(), b":", x))


def encode_list(x, r):
    r.append(b"l")
    for i in x:
        encode_func[type(i)](i, r)
    r.append(b"e")


def encode_dict(x, r):
    r.append(b"d")
    for k, v in sorted(x.items()):
        r.extend((str(len(k)).encode(), b":", k))
        encode_func[type(v)](v, r)
    r.append(b"e")


encode_func = {
    Bencached: encode_bencached,
    int: encode_int,
    str: encode_string,
    bytes: encode_string,
    list: encode_list,
    tuple: encode_list,
    dict: encode_dict
}

try:
    from types import BooleanType

    encode_func[BooleanType] = encode_bool
except ImportError:
    pass


def bencode(x):
    r = []
    encode_func[type(x)](x, r)
    return b"".join(r)



# encoding: utf-8

import socket
from collections import deque
from hashlib import sha1
from random import randint
from socket import inet_ntoa
from struct import unpack
from threading import Timer, Thread, Lock
from time import sleep

from notedrive.magnet.dht.bencode import bencode, bdecode

# 一些公共节点地址，用来本地爬虫加入dht网络
BOOTSTRAP_NODES = {
    ("router.bittorrent.com", 6881),
    ("dht.transmissionbt.com", 6881),
    ("router.utorrent.com", 6881)
}

# 一些全局变量
TID_LENGTH = 2
RE_JOIN_DHT_INTERVAL = 3
TOKEN_LENGTH = 2


# 用来生成随机字串
def entropy(length):
    return "".join(chr(randint(0, 255)) for _ in range(length))


# 根据随机字串生成nid，这样的话，自然很可能和其他的节点ID重复
def random_id():
    h = sha1()
    # h.update(entropy(20))
    return h.digest()


# 解析node，node长度为26，其中20位为nid，4位为ip，2位为port
def decode_nodes(nodes):
    n = []
    length = len(nodes)
    if (length % 26) != 0:
        return n
    for i in range(0, length, 26):
        nid = nodes[i:i + 20]
        ip = inet_ntoa(nodes[i + 20:i + 24])
        port = unpack("!H", nodes[i + 24:i + 26])[0]
        n.append((nid, ip, port))
    return n


# 定时器函数
def timer(t, f):
    Timer(t, f).start()


# 获取“邻居”nid
def get_neighbor(target, end=10):
    return target[:end] + random_id()[end:]


# 一个node的结构
class KNode():
    def __init__(self, nid, ip, port):
        self.nid = nid
        self.ip = ip
        self.port = port


# DHT类，继承自Thread类
class DHT(Thread):
    def __init__(self, master, port):
        Thread.__init__(self)

        # 用来输出和保存种子hash的对象
        self.master = master

        # 创建队列，相当于路由表的简单模拟
        # 根据原作者的注释来看，队列的最大长度越大，速度越快，耗费带宽越大
        self.nodes = deque(maxlen=2000)

        # 生成的自身的网络节点
        self.nid = random_id()

        # 创建socket
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
        self.socket.bind(("0.0.0.0", port))

        # 处理"get_peers"和"announce_peer"请求的函数
        self.process_request_actions = {
            "get_peers": self.on_get_peers_request,
            "announce_peer": self.on_announce_peer_request,
        }

    # 发包函数，发送的数据需要使用bencode编码
    def send_krpc(self, msg, address):
        try:
            self.socket.sendto(bencode(msg), address)
        except:
            pass

    # find node 方法，用来查找对应的node
    def find_nodes(self, address, nid=None):
        if not nid:
            nid = self.nid
        # 扩大寻找node的范围
        nid = get_neighbor(nid)
        tid = entropy(TID_LENGTH)
        msg = {
            "t": tid,
            "y": "q",
            "q": "find_node",
            "a": {"id": nid, "target": random_id()}
        }

        self.send_krpc(msg, address)

    # 加入dht网络
    def joinDHT(self):
        for address in BOOTSTRAP_NODES:
            self.find_nodes(address)

    # 重新加入dht网络
    def re_join_DHT(self):
        if len(self.nodes) == 0:
            self.joinDHT()
        timer(RE_JOIN_DHT_INTERVAL, self.re_join_DHT)

    # 重写run方法
    def run(self):
        self.re_join_DHT()
        while 1:
            try:
                # 接收数据并解码
                (data, address) = self.socket.recvfrom(65536)
                msg = bdecode(data)
                self.on_message(msg, address)
            except:
                pass

    # 处理find node方法的请求
    def process_find_node_response(self, msg, address):
        # 将node解码
        nodes = decode_nodes(msg["r"]["nodes"])
        for node in nodes:
            (nid, ip, port) = node
            if len(nid) != 20:
                continue
            if ip == "0.0.0.0":
                continue
            n = KNode(nid, ip, port)
            # 添加到node队列里
            self.nodes.append(n)

    # 发送get peers 的请求，在msg参数中包含种子hash的信息
    def on_get_peers_request(self, msg, address):
        try:
            # 获取hash值
            infohash = msg["a"]["info_hash"]
            # 输出并保存
            self.master.log(infohash, "get_peers")
            tid = msg["t"]
            nid = msg["a"]["id"]
            token = infohash[:TOKEN_LENGTH]
            msg = {
                "t": tid,
                "y": "r",
                "r": {
                    "id": get_neighbor(infohash, self.nid),
                    "nodes": "",
                    "token": token
                }
            }
            self.send_krpc(msg, address)
        except:
            pass

    # 发送announce peer的请求，同样在msg参数中包含种子hash信息
    def on_announce_peer_request(self, msg, address):
        try:
            infohash = msg['a']['info_hash']
            self.master.log(infohash, "announce")
        except:
            print("error in on_announce_peer_request")

        finally:
            self.ok(msg, address)

    # 对announce peer请求的响应
    def ok(self, msg, address):
        try:
            tid = msg["t"]
            nid = msg["a"]["id"]
            msg = {
                "t": tid,
                "y": "r",
                "r": {
                    "id": get_neighbor(nid, self.nid)
                }
            }
            self.send_krpc(msg, address)
        except:
            print("error in ok")

    # 处理接收到的消息
    def on_message(self, msg, address):
        try:
            # 处理response消息
            if msg["y"] == "r":
                if msg["r"].has_key("nodes"):
                    self.process_find_node_response(msg, address)
            # 处理request消息
            elif msg["y"] == "q":
                try:
                    self.process_request_actions[msg["q"]](msg, address)
                except:
                    self.play_dead(msg, address)
        except:
            print
            "error in on_message"

    # 发送错误消息
    def play_dead(self, msg, address):
        try:
            tid = msg["t"]
            msg = {
                "t": tid,
                "y": "e",
                "e": [202, "Server Error"]
            }
            self.send_krpc(msg, address)
        except:
            pass

    # 对队列里的node继续发送find node请求，获取更多的node
    def auto_send_find_node(self):
        wait = 1.0 / 2000
        while 1:
            try:
                node = self.nodes.popleft()
                self.find_nodes((node.ip, node.port), node.nid)
            except:
                pass
            sleep(wait)


# 用来输出及保存种子hash
class Master():
    def __init__(self):
        self.mutex = Lock()

    def log(self, infohash, source):
        global inum
        infohash = infohash.encode("hex")
        inum += 1
        print(inum, infohash, source)

        file = open("hash.log", "a")
        file.write(str(infohash) + "n")
        file.close()


inum = 0
t1 = DHT(Master(), 8006)
t1.start()
t1.auto_send_find_node()



from enum import Enum


class FieldType(bytes, Enum):
    NONE = (0, ['none', 'NONE'])
    STRING = (10, ['string', 'STRING', 'str', 'STR'])
    VARCHAR = (11, ['varchar', 'VARCHAR'])
    FLOAT = (20, ['float', 'FLOAT'])
    DOUBLE = (30, ['double', 'DOUBLE'])
    BIGINT = (40, ['bigint', 'BIGINT'])
    INTEGER = (50, ['integer', 'int', 'INTEGER', 'INT'])
    TINYINT = (51, ['tinyint', 'TINYINT'])

    def __init__(self, value, names):
        self._value_ = value
        self.names = names

    def __new__(cls, value, names):
        obj = bytes.__new__(cls, [value])
        obj._value_ = value
        obj.names = names
        return obj

    @staticmethod
    def parse(field_type):
        if isinstance(field_type, FieldType):
            return field_type
        for filed in FieldType:
            if field_type in filed.names:
                return filed
        return FieldType.NONE


class Field:
    def __init__(self, name, field_type, comment=""):
        self.name = name
        self.field_type = FieldType.parse(field_type)
        self.comment = comment

    @staticmethod
    def double_field(name, comment=""):
        return Field(name, FieldType.DOUBLE, comment)

    @staticmethod
    def bigint_field(name, comment=""):
        return Field(name, FieldType.BIGINT, comment)

    @staticmethod
    def float_field(name, comment=""):
        return Field(name, FieldType.FLOAT, comment)

    @staticmethod
    def string_field(name, comment=""):
        return Field(name, FieldType.STRING, comment)

    @staticmethod
    def varchar_field(name, comment=""):
        return Field(name, FieldType.VARCHAR, comment)

    @staticmethod
    def integer_field(name, comment=""):
        return Field(name, FieldType.INTEGER, comment)



import base64
import getpass
import http.client
import itertools
import json
import logging
import math
import os
import ssl
import string
import subprocess
import sys
import time
import urllib
import urllib.parse
import urllib.request

# -------------------------------- Constants --------------------------------- #

DEFAULT_PORT = 6800
SERVER_URI_FORMAT = '{}://{}:{:d}/jsonrpc'

# Status values for unfinished downloads.
TEMPORARY_STATUS = ('active', 'waiting', 'paused')
# Status values for finished downloads.
FINAL_STATUS = ('complete', 'error')

ARIA2_CONTROL_FILE_EXT = '.aria2'

# Encoding to use when inserting bytes objects into JSON.
JSON_ENCODING = 'utf-8'

# Module logger.
LOGGER = logging.getLogger(name=__name__)


# -------------------------- Convenience Functions --------------------------- #

def to_json_list(objs):
    """
    Wrap strings in lists. Other iterables are converted to lists directly.
    """
    if isinstance(objs, str):
        return [objs]
    if not isinstance(objs, list):
        return list(objs)
    return objs


def add_options_and_position(params, options=None, position=None):
    """
    Convenience method for adding options and position to parameters.
    """
    if options:
        params.append(options)
    if position:
        if not isinstance(position, int):
            try:
                position = int(position)
            except ValueError:
                position = -1
        if position >= 0:
            params.append(position)
    return params


def get_status(response):
    """
    Process a status response.
    """
    if response:
        try:
            return response['status']
        except KeyError:
            LOGGER.error('no status returned from Aria2 RPC server')
            return 'error'
    else:
        LOGGER.error('no response from server')
        return 'error'


def random_token(length, valid_chars=None):
    """
    Get a random secret token for the Aria2 RPC server.

    length:
      The length of the token

    valid_chars:
      A list or other ordered and indexable iterable of valid characters. If not
      given of None, asciinumberic characters with some punctuation characters
      will be used.
    """
    if not valid_chars:
        valid_chars = string.ascii_letters + string.digits + '!@#$%^&*()-_=+'
    number_of_chars = len(valid_chars)
    bytes_to_read = math.ceil(math.log(number_of_chars) / math.log(0x100))
    max_value = 0x100 ** bytes_to_read
    max_index = number_of_chars - 1
    token = ''
    for _ in range(length):
        value = int.from_bytes(os.urandom(bytes_to_read), byteorder='little')
        index = round((value * max_index) / max_value)
        token += valid_chars[index]
    return token


# ---------------- From python3-aur's ThreadedServers.common ----------------- #

def format_bytes(size):
    """
    Convert bytes to a human-friendly units..
    """
    if size < 0x400:
        return '{:d} B'.format(size)
    size = float(size) / 0x400
    for prefix in ('KiB', 'MiB', 'GiB', 'TiB', 'PiB', 'EiB', 'ZiB'):
        if size < 0x400:
            return '{:0.02f} {}'.format(size, prefix)
        size /= 0x400
    return '{:0.02f} YiB'.format(size)


def format_seconds(seconds):
    """
    Convert seconds to hours, minutes and seconds.
    """
    time_str = ''
    for base, char in (
            (60, 's'),
            (60, 'm'),
            (24, 'h')
    ):
        seconds, remainder = divmod(seconds, base)
        if seconds == 0:
            return '{:d}{}{}'.format(remainder, char, time_str)
        if remainder != 0:
            time_str = '{:02d}{}{}'.format(remainder, char, time_str)
    return '{:d}d{}'.format(seconds, string)


# --------------------------------- FakeLock --------------------------------- #

class FakeLock:
    """
    Dummy context manager to be used as a lock placeholder.
    """

    def __enter__(self):
        return None

    def __exit__(self, exc_type, exc_value, traceback):
        return


# ---------------------------- Aria2JsonRpcError ----------------------------- #

class Aria2JsonRpcError(Exception):
    """
    Base exception raised by this module.
    """

    def __init__(self, msg, connection_error=False):
        super(Aria2JsonRpcError, self).__init__(self, msg)
        self.msg = msg
        self.connection_error = connection_error

    def __str__(self):
        return '{}: {}'.format(self.__class__.__name__, self.msg)


# ------------------------------ ServerInstance ------------------------------ #

class Aria2RpcServer(object):
    """
    Wrapper to manage starting and stopping an Aria2 RPC server within a context
    manager.
    """

    def __init__(self, cmd, token, port, identity, a2jr_kwargs=None, timeout=10, scheme='http', host='localhost',
                 nice=True, lock=None, quiet=True):
        """
        cmd:
          A list representing the aria2c command to be run along with additional
          arguments, eg.

              ['aria2c', '--rpc-listen-all=false', '--continue']

          or

              ['sudo', 'aria2c', '--rpc-listen-all=false', '--continue']

          This MUST NOT contain any of the following options:

              --enable-rpc
              --rpc-secret
              --rpc-listen-port

          For local use it is recommended to include "--rpc-listen-all=false".

        token:
          The RPC token to pass to aria2c's --rpc-secret option.

        port:
          The RPC listen port to pass to aria2c's --rpc-listen-port option.

        identity:
          The identity to use in the Aria2JsonRpc returned by __enter__.

        args:
          Additional arguments to aria2c. These must not include --rpc-secret or
          --rpc-listen-port.

        a2jr_kwargs:
          Keyword arguments to pass to Aria2JsonRpc.

        timeout:
          The timeout when waiting for the server to shut down.

        nice:
          Default value of "nice" argument to kill() method.

        lock:
          An optional threading/multiprocessing lock object that can be used as a
          context manager to lock the launch method when starting and stopping the
          server.

        quiet:
          If True, suppress output from the RPC server if one is launched.
        """
        self.cmd = cmd
        self.cmd.append('--stop-with-process={:d}'.format(os.getpid()))
        self.token = token
        self.port = port
        self.identity = identity

        if a2jr_kwargs is None:
            a2jr_kwargs = dict()
        a2jr_kwargs['token'] = token
        a2jr_kwargs['setup_function'] = self.launch

        self.a2jr_kwargs = a2jr_kwargs

        self.timeout = timeout
        self.scheme = scheme
        self.host = host

        self.cmd.extend((
            '--enable-rpc',
            '--rpc-secret', token,
            '--rpc-listen-port', str(port)
        ))

        self.process = None
        self.a2jr = None
        self.nice = nice
        if lock is None:
            lock = FakeLock()
        self.lock = lock
        self.quiet = quiet

    def get_a2jr(self, new=False):
        """
        Get an instance of Aria2JsonRpc to interface with the server. This instance
        will have its setup_function attribute set so that it can launch the server
        dynamically before the first request is sent.
        """
        if self.a2jr is None or new:
            uri = SERVER_URI_FORMAT.format(self.scheme, self.host, self.port)
            self.a2jr = Aria2JsonRpc(self.identity, uri, **self.a2jr_kwargs)
        return self.a2jr

    def kill(self, nice=None):
        """
        Stop the Aria2c RPC server launched by this object if there is one.

        nice:
          If True, use the RPC "shutdown" method before trying the "forceShutdown"
          method. This waits up to 3 seconds even if there are no current downloads.
        """
        if self.process is None:
            return True

        with self.lock:
            LOGGER.debug('{}: attempting to stop PID {:d}'.format(
                self.__class__.__name__,
                self.process.pid
            ))

            # Try to kill the server with increasing insistence:
            # * RPC shutdown method
            # * RPC forceShutdown method
            # * SIGTERM
            # * SIGKILL
            a2jr = self.get_a2jr()
            if nice is None:
                nice = self.nice
            if nice:
                methods = (
                    a2jr.shutdown,
                    a2jr.forceShutdown,
                    self.process.terminate,
                    self.process.kill
                )
            else:
                methods = (
                    self.process.terminate,
                    self.process.kill
                )
            for schwarzenegger in methods:
                try:
                    schwarzenegger()
                except Aria2JsonRpcError:
                    pass
                try:
                    exit_code = self.process.wait(self.timeout)
                except subprocess.TimeoutExpired:
                    exit_code = None
                if exit_code is not None:
                    LOGGER.debug('{}: PID {:d} exit code: {:d}'.format(
                        self.__class__.__name__,
                        self.process.pid,
                        exit_code
                    ))
                    self.process = None
                    return True

        LOGGER.error('{}: failed to kill PID {:d}'.format(
            self.__class__.__name__,
            self.process.pid
        ))
        return False

    def launch(self):
        """
        Launch an instance of the server if necessary and return an Aria2JsonRpc
        object to interface with it.
        """
        with self.lock:
            a2jr = self.get_a2jr(new=True)
            a2jr.setup_function = None
            # Check if the server is already running.
            try:
                a2jr.getVersion()
            except Aria2JsonRpcError as err:
                if not err.connection_error:
                    raise err
                launch = True
            else:
                launch = False

            if launch:
                LOGGER.debug(
                    '%s: no response to version request from server: %s',
                    self.__class__.__name__,
                    self.cmd
                )
                self.kill()
                LOGGER.debug(
                    '%s: attempting to launch Aria2 RPC server: %s',
                    self.__class__.__name__,
                    self.cmd
                )
                if self.quiet:
                    self.process = subprocess.Popen(
                        self.cmd,
                        stdout=subprocess.DEVNULL,
                        stderr=subprocess.DEVNULL
                    )
                else:
                    self.process = subprocess.Popen(
                        self.cmd,
                        stdout=sys.stderr
                    )
            timeout = time.time() + self.timeout
            # Wait for the server to start listening.
            while True:
                try:
                    self.a2jr.getVersion()
                except Aria2JsonRpcError as err:
                    if err.connection_error:
                        time.sleep(0.25)
                        if time.time() > timeout:
                            break
                    else:
                        raise err
                else:
                    break
        return self.a2jr

    def __enter__(self):
        """
        Same as launch().
        """
        return self.launch()

    def __exit__(self, exc_type, exc_value, traceback):
        """
        Invokes kill()
        """
        self.kill()


# ---------------------------- Aria2JsonRpc Class ---------------------------- #

class Aria2JsonRpc(object):
    """
    Interface class for interacting with an Aria2 RPC server.
    """

    # TODO: certificate options, etc.

    def __init__(self, identity, uri,
                 mode='normal',
                 token=None,
                 http_user=None, http_passwd=None,
                 server_cert=None, client_cert=None, client_cert_password=None,
                 ssl_protocol=None,
                 setup_function=None
                 ):
        """
        identity: the identity to send to the RPC interface

        uri: the URI of the RPC interface

        mode:
          normal - process requests immediately
          batch - queue requests (run with "process_queue")
          format - return RPC request objects

        token:
          RPC method-level authorization token (set using `--rpc-secret`)

        http_user, http_password:
          HTTP Basic authentication credentials (deprecated)

        server_cert:
          server certificate for HTTPS connections

        client_cert:
          client certificate for HTTPS connections

        client_cert_password:
          prompt for client certificate password

        ssl_protocol:
          SSL protocol from the ssl module

        setup_function:
          A function to invoke prior to the first server call. This could be the
          launch() method of an Aria2RpcServer instance, for example. This attribute
          is set automatically in instances returned from Aria2RpcServer.get_a2jr()
        """
        self.identity = identity
        self.uri = uri
        self.mode = mode
        self.queue = []
        self.handlers = dict()
        self.token = token
        self.setup_function = setup_function

        if None not in (http_user, http_passwd):
            self.add_HTTPBasicAuthHandler(http_user, http_passwd)

        if server_cert or client_cert:
            self.add_HTTPSHandler(
                server_cert=server_cert,
                client_cert=client_cert,
                client_cert_password=client_cert_password,
                protocol=ssl_protocol
            )

        self.update_opener()

    def iter_handlers(self):
        """
        Iterate over handlers.
        """
        for name in ('HTTPS', 'HTTPBasicAuth'):
            try:
                yield self.handlers[name]
            except KeyError:
                pass

    def update_opener(self):
        """
        Build an opener from the current handlers.
        """
        self.opener = urllib.request.build_opener(*self.iter_handlers())

    def remove_handler(self, name):
        """
        Remove a handler.
        """
        try:
            del self.handlers[name]
        except KeyError:
            pass

    def add_HTTPBasicAuthHandler(self, user, passwd):
        """
        Add a handler for HTTP Basic authentication.

        If either user or passwd are None, the handler is removed.
        """
        handler = urllib.request.HTTPBasicAuthHandler()
        handler.add_password(
            realm='aria2',
            uri=self.uri,
            user=user,
            passwd=passwd,
        )
        self.handlers['HTTPBasicAuth'] = handler

    def remove_HTTPBasicAuthHandler(self):
        """
        Remove the HTTP Basic authentication handler.
        """
        self.remove_handler('HTTPBasicAuth')

    def add_HTTPSHandler(self,
                         server_cert=None,
                         client_cert=None,
                         client_cert_password=None,
                         protocol=None,
                         ):
        """
        Add a handler for HTTPS connections with optional server and client
        certificates.
        """
        if not protocol:
            protocol = ssl.PROTOCOL_TLSv1
        #       protocol = ssl.PROTOCOL_TLSv1_1 # openssl 1.0.1+
        #       protocol = ssl.PROTOCOL_TLSv1_2 # Python 3.4+
        context = ssl.SSLContext(protocol)

        if server_cert:
            context.verify_mode = ssl.CERT_REQUIRED
            context.load_verify_locations(cafile=server_cert)
        else:
            context.verify_mode = ssl.CERT_OPTIONAL

        if client_cert:
            context.load_cert_chain(client_cert, password=client_cert_password)

        self.handlers['HTTPS'] = urllib.request.HTTPSHandler(
            context=context,
            check_hostname=False
        )

    def remove_HTTPSHandler(self):
        """
        Remove the HTTPS handler.
        """
        self.remove_handler('HTTPS')

    def send_request(self, req_obj):
        """
        Send the request and return the response.
        """
        if self.setup_function:
            self.setup_function()
            self.setup_function = None
        LOGGER.debug(json.dumps(req_obj, indent='  ', sort_keys=True))
        req = json.dumps(req_obj).encode('UTF-8')
        try:
            with self.opener.open(self.uri, req) as handle:
                obj = json.loads(handle.read().decode())
                try:
                    return obj['result']
                except KeyError:
                    raise Aria2JsonRpcError('unexpected result: {}'.format(obj))
        except (urllib.error.URLError, ConnectionResetError) as err:
            # This should work but URLError does not set the errno attribute:
            # e.errno == errno.ECONNREFUSED
            err_str = str(err)
            raise Aria2JsonRpcError(
                err_str,
                connection_error=(
                        '111' in err_str or isinstance(err, ConnectionResetError)
                )
            )
        except http.client.BadStatusLine as err:
            raise Aria2JsonRpcError('{}: BadStatusLine: {} (HTTPS error?)'.format(
                self.__class__.__name__, err
            ))

    def jsonrpc(self, method, params=None, prefix='aria2.'):
        """
        POST a request to the RPC interface.
        """
        if not params:
            params = []

        if self.token is not None:
            token_str = 'token:{}'.format(self.token)
            if method == 'multicall':
                for param in params[0]:
                    try:
                        param['params'].insert(0, token_str)
                    except KeyError:
                        param['params'] = [token_str]
            else:
                params.insert(0, token_str)

        req_obj = {
            'jsonrpc': '2.0',
            'id': self.identity,
            'method': prefix + method,
            'params': params,
        }
        if self.mode == 'batch':
            self.queue.append(req_obj)
            return None
        if self.mode == 'format':
            return req_obj
        return self.send_request(req_obj)

    def process_queue(self):
        """
        Processed queued requests.
        """
        req_obj = self.queue
        self.queue = []
        return self.send_request(req_obj)

    # ----------------------------- Standard Methods ----------------------------- #

    def addUri(self, uris, options=None, position=None):
        """
        aria2.addUri method

        uris: list of URIs

        options: dictionary of additional options

        position: position in queue

        Returns a GID
        """
        params = [uris]
        params = add_options_and_position(params, options, position)
        return self.jsonrpc('addUri', params)

    def addTorrent(self, torrent, uris=None, options=None, position=None):
        """
        aria2.addTorrent method

        torrent: base64-encoded torrent file

        uris: list of webseed URIs

        options: dictionary of additional options

        position: position in queue

        Returns a GID.
        """
        params = [torrent]
        if uris:
            params.append(uris)
        params = add_options_and_position(params, options, position)
        return self.jsonrpc('addTorrent', params)

    def addMetalink(self, metalink, options=None, position=None):
        """
        aria2.addMetalink method

        metalink: base64-encoded metalink file

        options: dictionary of additional options

        position: position in queue

        Returns an array of GIDs.
        """
        params = [metalink]
        params = add_options_and_position(params, options, position)
        return self.jsonrpc('addTorrent', params)

    def remove(self, gid):
        """
        aria2.remove method

        gid: GID to remove
        """
        params = [gid]
        return self.jsonrpc('remove', params)

    def forceRemove(self, gid):
        """
        aria2.forceRemove method

        gid: GID to remove
        """
        params = [gid]
        return self.jsonrpc('forceRemove', params)

    def pause(self, gid):
        """
        aria2.pause method

        gid: GID to pause
        """
        params = [gid]
        return self.jsonrpc('pause', params)

    def pauseAll(self):
        """
        aria2.pauseAll method
        """
        return self.jsonrpc('pauseAll')

    def forcePause(self, gid):
        """
        aria2.forcePause method

        gid: GID to pause
        """
        params = [gid]
        return self.jsonrpc('forcePause', params)

    def forcePauseAll(self):
        """
        aria2.forcePauseAll method
        """
        return self.jsonrpc('forcePauseAll')

    def unpause(self, gid):
        """
        aria2.unpause method

        gid: GID to unpause
        """
        params = [gid]
        return self.jsonrpc('unpause', params)

    def unpauseAll(self):
        """
        aria2.unpauseAll method
        """
        return self.jsonrpc('unpauseAll')

    def tellStatus(self, gid, keys=None):
        """
        aria2.tellStatus method

        gid: GID to query

        keys: subset of status keys to return (all keys are returned otherwise)

        Returns a dictionary.
        """
        params = [gid]
        if keys:
            params.append(keys)
        return self.jsonrpc('tellStatus', params)

    def getUris(self, gid):
        """
        aria2.getUris method

        gid: GID to query

        Returns a list of dictionaries.
        """
        params = [gid]
        return self.jsonrpc('getUris', params)

    def getFiles(self, gid):
        """
        aria2.getFiles method

        gid: GID to query

        Returns a list of dictionaries.
        """
        params = [gid]
        return self.jsonrpc('getFiles', params)

    def getPeers(self, gid):
        """
        aria2.getPeers method

        gid: GID to query

        Returns a list of dictionaries.
        """
        params = [gid]
        return self.jsonrpc('getPeers', params)

    def getServers(self, gid):
        """
        aria2.getServers method

        gid: GID to query

        Returns a list of dictionaries.
        """
        params = [gid]
        return self.jsonrpc('getServers', params)

    def tellActive(self, keys=None):
        """
        aria2.tellActive method

        keys: same as tellStatus

        Returns a list of dictionaries. The dictionaries are the same as those
        returned by tellStatus.
        """
        if keys:
            params = [keys]
        else:
            params = None
        return self.jsonrpc('tellActive', params)

    def tellWaiting(self, offset, num, keys=None):
        """
        aria2.tellWaiting method

        offset: offset from start of waiting download queue
                (negative values are counted from the end of the queue)

        num: number of downloads to return

        keys: same as tellStatus

        Returns a list of dictionaries. The dictionaries are the same as those
        returned by tellStatus.
        """
        params = [offset, num]
        if keys:
            params.append(keys)
        return self.jsonrpc('tellWaiting', params)

    def tellStopped(self, offset, num, keys=None):
        """
        aria2.tellStopped method

        offset: offset from oldest download (same semantics as tellWaiting)

        num: same as tellWaiting

        keys: same as tellStatus

        Returns a list of dictionaries. The dictionaries are the same as those
        returned by tellStatus.
        """
        params = [offset, num]
        if keys:
            params.append(keys)
        return self.jsonrpc('tellStopped', params)

    def changePosition(self, gid, pos, how):
        """
        aria2.changePosition method

        gid: GID to change

        pos: the position

        how: "POS_SET", "POS_CUR" or "POS_END"
        """
        params = [gid, pos, how]
        return self.jsonrpc('changePosition', params)

    def changeUri(self, gid, fileIndex, delUris, addUris, position=None):
        """
        aria2.changePosition method

        gid: GID to change

        fileIndex: file to affect (1-based)

        delUris: URIs to remove

        addUris: URIs to add

        position: where URIs are inserted, after URIs have been removed
        """
        params = [gid, fileIndex, delUris, addUris]
        if position:
            params.append(position)
        return self.jsonrpc('changePosition', params)

    def getOption(self, gid):
        """
        aria2.getOption method

        gid: GID to query

        Returns a dictionary of options.
        """
        params = [gid]
        return self.jsonrpc('getOption', params)

    def changeOption(self, gid, options):
        """
        aria2.changeOption method

        gid: GID to change

        options: dictionary of new options
                 (not all options can be changed for active downloads)
        """
        params = [gid, options]
        return self.jsonrpc('changeOption', params)

    def getGlobalOption(self):
        """
        aria2.getGlobalOption method

        Returns a dictionary.
        """
        return self.jsonrpc('getGlobalOption')

    def changeGlobalOption(self, options):
        """
        aria2.changeGlobalOption method

        options: dictionary of new options
        """
        params = [options]
        return self.jsonrpc('changeGlobalOption', params)

    def getGlobalStat(self):
        """
        aria2.getGlobalStat method

        Returns a dictionary.
        """
        return self.jsonrpc('getGlobalStat')

    def purgeDownloadResult(self):
        """
        aria2.purgeDownloadResult method
        """
        self.jsonrpc('purgeDownloadResult')

    def removeDownloadResult(self, gid):
        """
        aria2.removeDownloadResult method

        gid: GID to remove
        """
        params = [gid]
        return self.jsonrpc('removeDownloadResult', params)

    def getVersion(self):
        """
        aria2.getVersion method

        Returns a dictionary.
        """
        return self.jsonrpc('getVersion')

    def getSessionInfo(self):
        """
        aria2.getSessionInfo method

        Returns a dictionary.
        """
        return self.jsonrpc('getSessionInfo')

    def shutdown(self):
        """
        aria2.shutdown method
        """
        return self.jsonrpc('shutdown')

    def forceShutdown(self):
        """
        aria2.forceShutdown method
        """
        return self.jsonrpc('forceShutdown')

    def multicall(self, methods):
        """
        aria2.multicall method

        methods: list of dictionaries (keys: methodName, params)

        The method names must be those used by Aria2c, e.g. "aria2.tellStatus".
        """
        return self.jsonrpc('multicall', [methods], prefix='system.')

    # --------------------------- Convenience Methods ---------------------------- #
    @staticmethod
    def b64encode_file(path):
        """
        Read a file into a base64-encoded string.
        """
        with open(path, 'rb') as handle:
            return base64.b64encode(handle.read()).encode(JSON_ENCODING)

    def add_torrent(self, path, uris=None, options=None, position=None):
        """
        A wrapper around addTorrent for loading files.
        """
        torrent = self.b64encode_file(path)
        return self.addTorrent(torrent, uris, options, position)

    def add_metalink(self, path, options=None, position=None):
        """
        A wrapper around addMetalink for loading files.
        """
        metalink = self.b64encode_file(path)
        return self.addMetalink(metalink, options, position)

    def get_status(self, gid):
        """
        Get the status of a single GID.
        """
        response = self.tellStatus(gid, ['status'])
        return get_status(response)

    def wait_for_final_status(self, gid, interval=1):
        """
        Wait for a GID to complete or fail and return its status.
        """
        if not interval or interval < 0:
            interval = 1
        while True:
            status = self.get_status(gid)
            if status in TEMPORARY_STATUS:
                time.sleep(interval)
            else:
                return status

    def get_statuses(self, gids):
        """
        Get the status of multiple GIDs. The status of each is yielded in order.
        """
        methods = [
            {
                'methodName': 'aria2.tellStatus',
                'params': [gid, ['gid', 'status']]
            }
            for gid in gids
        ]
        results = self.multicall(methods)
        if results:
            status = dict((r[0]['gid'], r[0]['status']) for r in results)
            for gid in gids:
                try:
                    yield status[gid]
                except KeyError:
                    LOGGER.error(
                        'Aria2 RPC server returned no status for GID %s',
                        gid
                    )
                    yield 'error'
        else:
            LOGGER.error('no response from Aria2 RPC server')
            for gid in gids:
                yield 'error'

    def wait_for_final_statuses(self, gids, interval=1):
        """
        Wait for multiple GIDs to complete or fail and return their statuses in
        order.

        gids:
          A flat list of GIDs.
        """
        if not interval or interval < 0:
            interval = 1
        statusmap = dict((gid, None) for gid in gids)
        remaining = list(
            gid for gid, s in statusmap.items() if s is None
        )
        while remaining:
            for gid, status in zip(remaining, self.get_statuses(remaining)):
                if status in TEMPORARY_STATUS:
                    continue
                statusmap[gid] = status
            remaining = list(
                gid for gid, status in statusmap.items() if status is None
            )
            if remaining:
                time.sleep(interval)
        for gid in gids:
            yield statusmap[gid]

    def print_global_status(self):
        """
        Print global status of the RPC server.
        """
        status = self.getGlobalStat()
        if status:
            numWaiting = int(status['numWaiting'])
            numStopped = int(status['numStopped'])
            keys = ['totalLength', 'completedLength']
            total = self.tellActive(keys)
            waiting = self.tellWaiting(0, numWaiting, keys)
            if waiting:
                total += waiting
            stopped = self.tellStopped(0, numStopped, keys)
            if stopped:
                total += stopped

            downloadSpeed = int(status['downloadSpeed'])
            uploadSpeed = int(status['uploadSpeed'])
            totalLength = sum(int(x['totalLength']) for x in total)
            completedLength = sum(int(x['completedLength']) for x in total)
            remaining = totalLength - completedLength

            status['downloadSpeed'] = format_bytes(downloadSpeed) + '/s'
            status['uploadSpeed'] = format_bytes(uploadSpeed) + '/s'

            preordered = ('downloadSpeed', 'uploadSpeed')

            rows = list(
                (k, v)
                for (k, v) in sorted(status.items())
                if k not in preordered
            )
            rows.extend((x, status[x]) for x in preordered)

            if totalLength > 0:
                rows.append(('total', format(format_bytes(totalLength))))
                rows.append(('completed', format(format_bytes(completedLength))))
                rows.append(('remaining', format(format_bytes(remaining))))
                if completedLength == totalLength:
                    eta = 'finished'
                else:
                    try:
                        eta = format_seconds(remaining // downloadSpeed)
                    except ZeroDivisionError:
                        eta = 'never'
                rows.append(('ETA', eta))

            width_l = max(len(r[0]) for r in rows)
            width_r = max(len(r[1]) for r in rows)
            width_r = max(width_r, len(self.uri) - (width_l + 2))
            fmt = '{{:<{:d}s}}  {{:>{:d}s}}'.format(width_l, width_r)

            print(self.uri)
            for row in rows:
                print(fmt.format(*row))

    def queue_uris(self, uris, options, interval=None):
        """
        Enqueue URIs and wait for download to finish while printing status at
        regular intervals.
        """
        gid = self.addUri(uris, options)
        print('GID: {}'.format(gid))

        if gid and interval is not None:
            blanker = ''
            while True:
                response = self.tellStatus(gid, ['status'])
                if response:
                    try:
                        status = response['status']
                    except KeyError:
                        print('error: no status returned from Aria2 RPC server')
                        break
                    print('{}\rstatus: {}'.format(blanker, status), end='')
                    blanker = ' ' * len(status)
                    if status in TEMPORARY_STATUS:
                        time.sleep(interval)
                    else:
                        break
                else:
                    print('error: no response from server')
                    break

    # ----------------------- Polymethod download handlers ----------------------- #

    def polymethod_enqueue_many(self, downloads):
        """
        Enqueue downloads.

        downloads: Same as polymethod_download().
        """
        methods = list(
            {
                'methodName': 'aria2.{}'.format(d[0]),
                'params': list(d[1:])
            } for d in downloads
        )
        return self.multicall(methods)

    def polymethod_wait_many(self, gid_lists, interval=1):
        """
        Wait for the GIDs to complete or fail and return their statuses.

        gids:
          A list of lists of GIDs.
        """
        # The flattened list of GIDs
        pending_gids = list(itertools.chain.from_iterable(gid_lists))
        statusmap = dict(zip(pending_gids, self.wait_for_final_statuses(pending_gids, interval=interval)))
        # Return the grouped statuses in order.
        for gids in gid_lists:
            yield list(statusmap.get(gid, 'error') for gid in gids)

    def polymethod_enqueue_one(self, download):
        """
        Same as polymethod_enqueue_many but for one element.
        """
        return getattr(self, download[0])(*download[1:])

    def polymethod_download(self, downloads, interval=1):
        """
        Enqueue a series of downloads and wait for them to finish. Iterate over the
        status of each, in order.

        downloads:
          An iterable over (<type>, <args>, ...) where <type> indicates the "add"
          method to use ('addUri', 'addTorrent', 'addMetalink') and everything that
          follows are arguments to pass to that method.

        interval:
          The status check interval while waiting.

        Iterates over the download status of finished downloads. "complete"
        indicates success. Lists of statuses will be returned for downloads that
        create multiple GIDs (e.g. metalinks).
        """
        gids = self.polymethod_enqueue_many(downloads)
        return self.polymethod_wait_many(gids, interval=interval)

    def polymethod_download_bool(self, *args, **kwargs):
        """
        A wrapper around polymethod_download() which returns a boolean for each
        download to indicate success (True) or failure (False).
        """
        #     for status in self.polymethod_download(*args, **kwargs):
        #       yield all(s == 'complete' for s in status)
        return list(all(s == 'complete' for s in status) for status in self.polymethod_download(*args, **kwargs))


# --------------------------------- argparse --------------------------------- #

def add_server_arguments(parser):
    """
    Common command-line arguments for the server.

    Accepts an argparse ArgumentParser or group.
    """
    parser.add_argument('-a', '--address', default='localhost', help='The server host. Default: %(default)s.')
    parser.add_argument('-p', '--port', type=int, default=DEFAULT_PORT, help='The server port. Default: %(default)s.')
    parser.add_argument('-s', '--scheme', default='http', help='The server scheme. Default: %(default)s.')

    # parser.add_argument('--auth', nargs=2, help='HTTP Basic authentication user and password.')

    parser.add_argument('--token', help='Secret RPC token.')

    parser.add_argument('--server-cert', help='HTTPS server certificate file, in PEM format.')
    parser.add_argument('--client-cert', help='HTTPS client certificate file, in PEM format.')
    parser.add_argument('--client-cert-password', help='Prompt for a client certificate password.')


def a2jr_from_args(identity, args):
    """
    Return a new Aria2JsonRpc object using the provided arguments.

    See `add_server_arguments`.
    """
    uri = SERVER_URI_FORMAT.format(args.scheme, args.address, args.port)

    #   if args.auth:
    #     http_user, http_passwd = args.auth
    #   else:
    #     http_user = None
    #     http_passwd = None

    token = args.token

    if args.client_cert and args.client_cert_password:
        client_cert_password = getpass.getpass('password for {}: '.format(args.client_cert))
    else:
        client_cert_password = None

    return Aria2JsonRpc(identity, uri, token=token,
                        #     http_user=http_user, http_passwd=http_passwd,
                        server_cert=args.server_cert,
                        client_cert=args.client_cert,
                        client_cert_password=client_cert_password,
                        )



import shutil
from concurrent.futures import ThreadPoolExecutor, as_completed
from random import random
from time import sleep
from typing import List

from .utils import *
from notesecret.secret import SecretManage
from requests_toolbelt import MultipartEncoder, MultipartEncoderMonitor
from tqdm import tqdm
from urllib3 import disable_warnings
from urllib3.exceptions import InsecureRequestWarning

# up_url='https://pc.woozooo.com/fileup.php'
up_url = 'http://up.woozooo.com/fileup.php'


def File(name='', id=0, time='', size='', _type='', downs=0, has_pwd=0, has_des=0, type='', ) -> dict:
    return {
        'id': id,
        'name': name,
        'size': size,
        'type': _type or type,
        'time': time,
        'downs': downs,
        'has_pwd': has_pwd,
        'has_des': has_des,
    }


def Folder(name='', id=0, has_pwd=False, desc='') -> dict:
    return {
        'name': name,
        'id': id,
        'has_pwd': has_pwd,
        'desc': desc,
    }


def FolderId(name, id) -> dict:
    return {
        'name': name,
        'id': id,
    }


def rec_file(name, id, _type='', size='', time='', type='') -> dict:
    return {
        'name': name,
        'id': id,
        'type': _type or type,
        'time': time,
        'size': size,
    }


def rec_folder(name, id, size, time, files) -> dict:
    return {
        'name': name,
        'id': id,
        'size': size,
        'time': time,
        'files': files,
    }


def file_detail(code=0, name='', size='', _type='', time='', desc='', pwd='', url='', durl='', type='') -> dict:
    return {
        'code': code,
        'name': name,
        'size': size,
        'type': _type or type,
        'time': time,
        'desc': desc,
        'pwd': pwd,
        'url': url,
        'durl': durl,
    }


def ShareInfo(code=0, name='', url='', pwd='', desc='', ) -> dict:
    return {
        'code': code,
        'name': name,
        'url': url,
        'pwd': pwd,
        'desc': desc
    }


def direct_url_info(code, name, durl) -> dict:
    return {
        'code': code,
        'name': name,
        'durl': durl
    }


def folder_info(name='', id=0, pwd='', time='', desc='', url='') -> dict:
    return {
        'name': name,
        'id': id,
        'time': time,
        'pwd': pwd,
        'desc': desc,
        'url': url
    }


def file_in_folder(name='', time='', size='', _type='', url='', type='') -> dict:
    return {
        'name': name,
        'time': time,
        'size': size,
        'type': _type or type,
        'url': url
    }


def FolderDetail(code=0, folder=None, files=None) -> dict:
    return {
        'code': code,
        'folder': folder,
        'files': files
    }


def find_filter(array: list, condition) -> list:
    """筛选出满足条件的 item
    condition(item) -> True
    """
    return [it for it in array if condition(it)]


def find_by_name(array: list, name: str):
    """使用文件名搜索(仅返回首个匹配项)"""
    for item in array:
        if name == item['name']:
            return item
    return None


def update_by_id(array: list, fid, **kwargs):
    """通过 id 搜索元素并更新"""
    item = find_by_id(array, fid)
    pos = array.index(item)
    array[pos] = array[pos].update(kwargs)


def find_by_id(array: list, fid: int):
    """使用 id 搜索(精确)"""
    for item in array:
        if fid == item['id']:
            return item
    return None


def pop_by_id(array: list, fid):
    for item in array:
        if item['id'] == fid:
            array.remove(item)
            return item
    return None


class CodeDetail:
    FAILED = -1
    SUCCESS = 0
    ID_ERROR = 1
    PASSWORD_ERROR = 2
    LACK_PASSWORD = 3
    ZIP_ERROR = 4
    MKDIR_ERROR = 5
    URL_INVALID = 6
    FILE_CANCELLED = 7
    PATH_ERROR = 8
    NETWORK_ERROR = 9
    CAPTCHA_ERROR = 10
    OFFICIAL_LIMITED = 11


class LanZouCloud(object):

    def __init__(self):
        self._session = requests.Session()
        self._captcha_handler = None
        self._limit_mode = True  # 是否保持官方限制
        self._timeout = 15  # 每个请求的超时(不包含下载响应体的用时)
        self._max_size = 100  # 单个文件大小上限 MB
        self._upload_delay = (0, 0)  # 文件上传延时
        #self._host_url = 'https://www.lanzous.com'
        self._host_url = 'https://pan.lanzoui.com'
        self._doupload_url = 'https://pc.woozooo.com/doupload.php'
        self._account_url = 'https://pc.woozooo.com/account.php'
        self._mydisk_url = 'https://pc.woozooo.com/mydisk.php'
        self._cookies = None
        self._headers = {
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/75.0.3770.100 Safari/537.36',
            # 'Referer': 'https://www.lanzous.com',
            # 'Referer': 'https://www.lanzoui.com',
            'Referer': 'https://pc.woozooo.com/mydisk.php',
            'Accept-Language': 'zh-CN,zh;q=0.9',  # 提取直连必需设置这个，否则拿不到数据
        }
        disable_warnings(InsecureRequestWarning)  # 全局禁用 SSL 警告

    def _get(self, url, **kwargs):
        kwargs.setdefault('timeout', self._timeout)
        kwargs.setdefault('headers', self._headers)
        for possiable_url in self.all_possiable_urls(url):
            try:
                return self._session.get(possiable_url, verify=False, **kwargs)
            except (ConnectionError, requests.RequestException):
                logger.warning(f"_get({possiable_url}) failed, try another domain")

        return None

    def _post(self, url, data, **kwargs):
        kwargs.setdefault('timeout', self._timeout)
        kwargs.setdefault('headers', self._headers)
        for possiable_url in self.all_possiable_urls(url):
            try:
                return self._session.post(possiable_url, data, verify=False, **kwargs)
            except (ConnectionError, requests.RequestException):
                logger.warning(f"_post({possiable_url}, {data}) failed, try another domain")

        return None

    def all_possiable_urls(self, lanzouyun_url: str) -> List[str]:
        if self._host_url not in lanzouyun_url:
            return [lanzouyun_url]

        old_domain = 'pan.lanzoui'

        base_host_urls = ['pan.lanzoux', 'wws.lanzoux', 'www.lanzoux', 'wwx.lanzoux',
                          'wws.lanzous', 'www.lanzous', 'wwx.lanzous', 'up.lanzoui'
                          ]

      
        return [lanzouyun_url, *[lanzouyun_url.replace(old_domain, base) for base in base_host_urls]]

    def ignore_limits(self):
        """解除官方限制"""
        logger.warning("*** You have enabled the big file upload and filename disguise features ***")
        logger.warning("*** This means that you fully understand what may happen and still agree to take the risk ***")
        self._limit_mode = False

    def set_max_size(self, max_size=100) -> int:
        """设置单文件大小限制(会员用户可超过 100M)"""
        if max_size < 100:
            return CodeDetail.FAILED
        self._max_size = max_size
        return CodeDetail.SUCCESS

    def set_upload_delay(self, t_range: tuple) -> int:
        """设置上传大文件数据块时，相邻两次上传之间的延时，减小被封号的可能"""
        if 0 <= t_range[0] <= t_range[1]:
            self._upload_delay = t_range
            return CodeDetail.SUCCESS
        return CodeDetail.FAILED

    def set_captcha_handler(self, captcha_handler):
        """设置下载验证码处理函数
        :param captcha_handler (img_data) -> str 参数为图片二进制数据,需返回验证码字符
        """
        self._captcha_handler = captcha_handler

    def login(self, username, password) -> int:
        """登录蓝奏云控制台"""
        login_data = {
            "action": "login",
            "task": "login",
            "setSessionId": "",
            "setToken": "",
            "setSig": "",
            "setScene": "",
            "username": username,
            "password": password
        }
        phone_header = {
            "User-Agent": "Mozilla/5.0 (Linux; Android 5.0; SM-G900P Build/LRX21T) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/82.0.4051.0 Mobile Safari/537.36"}
        html = self._get(self._account_url)
        if not html:
            return CodeDetail.NETWORK_ERROR
        formhash = re.findall(r'name="formhash" value="(.+?)"', html.text)
        if not formhash:
            return CodeDetail.FAILED
        login_data['formhash'] = formhash[0]
        html = self._post(self._account_url, login_data, headers=phone_header)
        if not html:
            return CodeDetail.NETWORK_ERROR
        if '登录成功' in html.text:
            self._cookies = html.cookies.get_dict()
            return CodeDetail.SUCCESS
        else:
            return CodeDetail.FAILED

    def get_cookie(self) -> dict:
        """获取用户 Cookie"""
        return self._cookies

    def login_by_cookie(self, cookie: dict = None, ylogin=None, phpdisk_info=None) -> int:
        """通过cookie登录"""
        secret = SecretManage()
        cookie = cookie or {
            'ylogin': secret.read("drive", "lanzou", 'ylogin', value=ylogin),
            'phpdisk_info': secret.read("drive", "lanzou", 'phpdisk_info', value=phpdisk_info)
        }

        self._session.cookies.update(cookie)
        html = self._get(self._account_url)
        if not html:
            return CodeDetail.NETWORK_ERROR
        return CodeDetail.FAILED if '网盘用户登录' in html.text else CodeDetail.SUCCESS

    def logout(self) -> int:
        """注销"""
        html = self._get(self._account_url, params={'action': 'logout'})
        if not html:
            return CodeDetail.NETWORK_ERROR
        return CodeDetail.SUCCESS if '退出系统成功' in html.text else CodeDetail.FAILED

    def delete(self, fid, is_file=True) -> int:
        """把网盘的文件、无子文件夹的文件夹放到回收站"""
        post_data = {'task': 6, 'file_id': fid} if is_file else {'task': 3, 'folder_id': fid}
        result = self._post(self._doupload_url, post_data)
        if not result:
            return CodeDetail.NETWORK_ERROR
        return CodeDetail.SUCCESS if result.json()['zt'] == 1 else CodeDetail.FAILED

    def clean_rec(self) -> int:
        """清空回收站"""
        post_data = {'action': 'delete_all', 'task': 'delete_all'}
        html = self._get(self._mydisk_url, params={'item': 'recycle', 'action': 'files'})
        if not html:
            return CodeDetail.NETWORK_ERROR
        post_data['formhash'] = re.findall(r'name="formhash" value="(.+?)"', html.text)[0]  # 设置表单 hash
        html = self._post(self._mydisk_url + '?item=recycle', post_data)
        if not html:
            return CodeDetail.NETWORK_ERROR
        return CodeDetail.SUCCESS if '清空回收站成功' in html.text else CodeDetail.FAILED

    def get_rec_dir_list(self) -> []:
        """获取回收站文件夹列表"""
        # 回收站中文件(夹)名只能显示前 17 个中文字符或者 34 个英文字符，如果这些字符相同，则在文件(夹)名后添加 (序号) ，以便区分
        html = self._get(self._mydisk_url, params={'item': 'recycle', 'action': 'files'})
        if not html:
            return []
        dirs = re.findall(r'folder_id=(\d+).+?>&nbsp;(.+?)\.{0,3}</a>.*\n+.*<td.+?>(.+?)</td>.*\n.*<td.+?>(.+?)</td>',
                          html.text)
        all_dir_list = []  # 文件夹信息列表
        dir_name_list = []  # 文件夹名列表d
        counter = 1  # 重复计数器
        for fid, name, size, time in dirs:
            if name in dir_name_list:  # 文件夹名前 17 个中文或 34 个英文重复
                counter += 1
                name = '{name}({counter})'.format(name=name, counter=counter)
            else:
                counter = 1
            dir_name_list.append(name)
            all_dir_list.append(rec_folder(name, int(fid), size, time, None))
        return all_dir_list

    def get_rec_file_list(self, folder_id=-1) -> list:
        """获取回收站文件列表"""
        if folder_id == -1:  # 列出回收站根目录文件
            # 回收站文件夹中的文件也会显示在根目录
            html = self._get(self._mydisk_url, params={'item': 'recycle', 'action': 'files'})
            if not html:
                return []
            html = remove_notes(html.text)
            files = re.findall(
                r'fl_sel_ids[^\n]+value="(\d+)".+?filetype/(\w+)\.gif.+?/>\s?(.+?)(?:\.{3})?</a>.+?<td.+?>([\d\-]+?)</td>',
                html, re.DOTALL)
            file_list = []
            file_name_list = []
            counter = 1
            for fid, ftype, name, time in sorted(files, key=lambda x: x[2]):
                if not name.endswith(ftype):  # 防止文件名太长导致丢失了文件后缀
                    name = name + '.' + ftype

                if name in file_name_list:  # 防止长文件名前 17:34 个字符相同重名
                    counter += 1
                    name = '{name}({counter})'.format(name=name, counter=counter)
                else:
                    counter = 1
                    file_name_list.append(name)
                file_list.append(rec_file(name, int(fid), ftype, size='', time=time))
            return file_list
        else:  # 列出回收站中文件夹内的文件,信息只有部分文件名和文件大小
            para = {'item': 'recycle', 'action': 'folder_restore', 'folder_id': folder_id}
            html = self._get(self._mydisk_url, params=para)
            if not html or '此文件夹没有包含文件' in html.text:
                return []
            html = remove_notes(html.text)
            files = re.findall(
                r'com/(\d+?)".+?filetype/(\w+)\.gif.+?/>&nbsp;(.+?)(?:\.{3})?</a> <font color="#CCCCCC">\((.+?)\)</font>',
                html)
            file_list = []
            file_name_list = []
            counter = 1
            for fid, ftype, name, size in sorted(files, key=lambda x: x[2]):
                if not name.endswith(ftype):  # 防止文件名太长丢失后缀
                    name = name + '.' + ftype
                if name in file_name_list:
                    counter += 1
                    name = '{name}({counter})'.format(name=name, counter=counter)  # 防止文件名太长且前17个字符重复
                else:
                    counter = 1
                    file_name_list.append(name)
                file_list.append(rec_file(name, int(fid), ftype, size=size, time=''))
            return file_list

    def get_rec_all(self):
        """获取整理后回收站的所有信息"""
        root_files = self.get_rec_file_list()  # 回收站根目录文件列表
        folder_list = []  # 保存整理后的文件夹列表
        for folder in self.get_rec_dir_list():  # 遍历所有子文件夹
            this_folder = rec_folder(folder['name'], folder['id'], folder['size'], folder['time'], [])
            for file in self.get_rec_file_list(folder['id']):  # 文件夹内的文件属性: name,id,type,size
                if find_by_id(root_files, file['id']):  # 根目录存在同名文件
                    file['time'] = pop_by_id(root_files, file['id'])['time']  # 从根目录删除, time 信息用来补充文件夹中的文件
                    this_folder['files'].append(file)
                else:  # 根目录没有同名文件(用户手动删了),文件还在文件夹中，只是根目录不显示，time 信息无法补全了
                    file['time'] = folder['time']  # 那就设置时间为文件夹的创建时间
                    this_folder['files'].append(file)
            folder_list.append(this_folder)
        return root_files, folder_list

    def _get_and_post(self, para, post_data):
        html = self._get(self._mydisk_url, params=para)
        if not html:
            return CodeDetail.NETWORK_ERROR
        # 此处的 formhash 与 login 时不同，不要尝试精简这一步
        post_data['formhash'] = re.findall(r'name="formhash" value="(\w+?)"', html.text)[0]  # 设置表单 hash
        html = self._post(self._mydisk_url + '?item=recycle', post_data)
        return html

    def delete_rec(self, fid, is_file=True) -> int:
        """彻底删除回收站文件(夹)"""
        # 彻底删除后需要 1.5s 才能调用 get_rec_file() ,否则信息没有刷新，被删掉的文件似乎仍然 "存在"
        if is_file:
            para = {'item': 'recycle', 'action': 'file_delete_complete', 'file_id': fid}
            post_data = {'action': 'file_delete_complete', 'task': 'file_delete_complete', 'file_id': fid}
        else:
            para = {'item': 'recycle', 'action': 'folder_delete_complete', 'folder_id': fid}
            post_data = {'action': 'folder_delete_complete', 'task': 'folder_delete_complete', 'folder_id': fid}

        # html = self._get(self._mydisk_url, params=para)
        # if not html:
        #     return CodeDetail.NETWORK_ERROR
        # # 此处的 formhash 与 login 时不同，不要尝试精简这一步
        # post_data['formhash'] = re.findall(r'name="formhash" value="(\w+?)"', html.text)[0]  # 设置表单 hash
        # html = self._post(self._mydisk_url + '?item=recycle', post_data)
        html = self._get_and_post(para, post_data)
        if not html:
            return CodeDetail.NETWORK_ERROR
        return CodeDetail.SUCCESS if '删除成功' in html.text else CodeDetail.FAILED

    def delete_rec_multi(self, *, files=None, folders=None) -> int:
        """彻底删除回收站多个文件(夹)
        :param files 文件 id 列表 List[int]
        :param folders 文件夹 id 列表 List[int]
        """
        if not files and not folders:
            return CodeDetail.FAILED
        para = {'item': 'recycle', 'action': 'files'}
        post_data = {'action': 'files', 'task': 'delete_complete_recycle'}
        if folders:
            post_data['fd_sel_ids[]'] = folders
        if files:
            post_data['fl_sel_ids[]'] = files
        html = self._get(self._mydisk_url, params=para)
        if not html:
            return CodeDetail.NETWORK_ERROR
        post_data['formhash'] = re.findall(r'name="formhash" value="(\w+?)"', html.text)[0]  # 设置表单 hash
        html = self._post(self._mydisk_url + '?item=recycle', post_data)
        if not html:
            return CodeDetail.NETWORK_ERROR
        return CodeDetail.SUCCESS if '删除成功' in html.text else CodeDetail.FAILED

    def recovery(self, fid, is_file=True) -> int:
        """从回收站恢复文件"""
        if is_file:
            para = {'item': 'recycle', 'action': 'file_restore', 'file_id': fid}
            post_data = {'action': 'file_restore', 'task': 'file_restore', 'file_id': fid}
        else:
            para = {'item': 'recycle', 'action': 'folder_restore', 'folder_id': fid}
            post_data = {'action': 'folder_restore', 'task': 'folder_restore', 'folder_id': fid}
        # html = self._get(self._mydisk_url, params=para)
        # if not html:
        #     return CodeDetail.NETWORK_ERROR
        # post_data['formhash'] = re.findall(r'name="formhash" value="(\w+?)"', html.text)[0]  # 设置表单 hash
        # html = self._post(self._mydisk_url + '?item=recycle', post_data)
        html = self._get_and_post(para, post_data)
        if not html:
            return CodeDetail.NETWORK_ERROR
        return CodeDetail.SUCCESS if '恢复成功' in html.text else CodeDetail.FAILED

    def recovery_multi(self, *, files=None, folders=None) -> int:
        """从回收站恢复多个文件(夹)"""
        if not files and not folders:
            return CodeDetail.FAILED
        para = {'item': 'recycle', 'action': 'files'}
        post_data = {'action': 'files', 'task': 'restore_recycle'}
        if folders:
            post_data['fd_sel_ids[]'] = folders
        if files:
            post_data['fl_sel_ids[]'] = files
        html = self._get(self._mydisk_url, params=para)
        if not html:
            return CodeDetail.NETWORK_ERROR
        post_data['formhash'] = re.findall(r'name="formhash" value="(.+?)"', html.text)[0]  # 设置表单 hash
        html = self._post(self._mydisk_url + '?item=recycle', post_data)
        if not html:
            return CodeDetail.NETWORK_ERROR
        return CodeDetail.SUCCESS if '恢复成功' in html.text else CodeDetail.FAILED

    def recovery_all(self) -> int:
        """从回收站恢复所有文件(夹)"""
        para = {'item': 'recycle', 'action': 'restore_all'}
        post_data = {'action': 'restore_all', 'task': 'restore_all'}
        # first_page = self._get(self._mydisk_url, params=para)
        # if not first_page:
        #     return CodeDetail.NETWORK_ERROR
        # post_data['formhash'] = re.findall(r'name="formhash" value="(.+?)"', first_page.text)[0]  # 设置表单 hash
        # second_page = self._post(self._mydisk_url + '?item=recycle', post_data)
        second_page = self._get_and_post(para, post_data)
        if not second_page:
            return CodeDetail.NETWORK_ERROR
        return CodeDetail.SUCCESS if '还原成功' in second_page.text else CodeDetail.FAILED

    def get_file_list(self, folder_id=-1) -> List[dict]:
        """获取文件列表"""
        page = 1
        file_list = []
        while True:
            post_data = {'task': 5, 'folder_id': folder_id, 'pg': page}
            resp = self._post(self._doupload_url, post_data)
            if not resp:  # 网络异常，重试
                continue
            else:
                resp = resp.json()
            if resp["info"] == 0:
                break  # 已经拿到了全部的文件信息
            else:
                page += 1  # 下一页
            # 文件信息处理
            for file in resp["text"]:
                file_list.append(File(
                    id=int(file['id']),
                    name=file['name_all'],
                    time=time_format(file['time']),  # 上传时间
                    size=file['size'],  # 文件大小
                    type=file['name_all'].split('.')[-1],  # 文件类型
                    downs=int(file['downs']),  # 下载次数
                    has_pwd=True if int(file['onof']) == 1 else False,  # 是否存在提取码
                    has_des=True if int(file['is_des']) == 1 else False  # 是否存在描述
                ))
        return file_list

    def get_dir_list(self, folder_id=-1) -> []:
        """获取子文件夹列表"""
        folder_list = []
        post_data = {'task': 47, 'folder_id': folder_id}
        resp = self._post(self._doupload_url, post_data)
        if not resp:
            return folder_list
        for folder in resp.json()['text']:
            folder_list.append(
                Folder(
                    id=int(folder['fol_id']),
                    name=folder['name'],
                    has_pwd=True if folder['onof'] == 1 else False,
                    desc=folder['folder_des'].strip('[]')
                ))
        return folder_list

    def clean_ghost_folders(self):
        """清除网盘中的幽灵文件夹"""

        # 可能有一些文件夹，网盘和回收站都看不见它，但是它确实存在，移动文件夹时才会显示
        # 如果不清理掉，不小心将文件移动进去就完蛋了
        def _clean(fid):
            for folder_ in self.get_dir_list(fid):
                real_folders.append(folder_)
                _clean(folder_['id'])

        folder_with_ghost = self.get_move_folders()
        pop_by_id(folder_with_ghost, -1)  # 忽视根目录
        real_folders = []
        _clean(-1)
        for folder in folder_with_ghost:
            if not find_by_id(real_folders, folder['id']):
                logger.debug("Delete ghost folder: {name} #{id}".format(name=folder['name'], id=folder['id']))
                if self.delete(folder['id'], False) != CodeDetail.SUCCESS:
                    return CodeDetail.FAILED
                if self.delete_rec(folder['id'], False) != CodeDetail.SUCCESS:
                    return CodeDetail.FAILED
        return CodeDetail.SUCCESS

    def get_full_path(self, folder_id=-1) -> []:
        """获取文件夹完整路径"""
        path_list = [FolderId('LanZouCloud', -1)]
        post_data = {'task': 47, 'folder_id': folder_id}
        resp = self._post(self._doupload_url, post_data)
        if not resp:
            return path_list
        for folder in resp.json()['info']:
            if folder['folderid'] and folder['name']:  # 有时会返回无效数据, 这两个字段中某个为 None
                path_list.append(FolderId(id=int(folder['folderid']), name=folder['name']))
        return path_list

    def _captcha_recognize(self, file_token):
        """识别下载时弹出的验证码,返回下载直链
        :param file_token 文件的标识码,每次刷新会变化
        """
        if not self._captcha_handler:  # 必需提前设置验证码处理函数
            logger.debug("Not set captcha handler function!")
            return None

        get_img_api = 'https://vip.d0.baidupan.com/file/imagecode.php?r=' + str(random())
        img_data = self._get(get_img_api).content
        captcha = self._captcha_handler(img_data)  # 用户手动识别验证码
        post_code_api = 'https://vip.d0.baidupan.com/file/ajax.php'
        post_data = {'file': file_token, 'bm': captcha}
        resp = self._post(post_code_api, post_data)
        if not resp or resp.json()['zt'] != 1:
            logger.debug("Captcha ERROR: {captcha}".format(captcha=captcha))
            return None
        logger.debug("Captcha PASS: {captcha}".format(captcha=captcha))
        return resp.json()['url']

    def get_file_info_by_url(self, share_url, pwd='') -> dict:
        """获取文件各种信息(包括下载直链)
        :param share_url: 文件分享链接
        :param pwd: 文件提取码(如果有的话)
        """
        if not is_file_url(share_url):  # 非文件链接返回错误
            return file_detail(CodeDetail.URL_INVALID, pwd=pwd, url=share_url)

        first_page = self._get(share_url)  # 文件分享页面(第一页)
        if not first_page:
            return file_detail(CodeDetail.NETWORK_ERROR, pwd=pwd, url=share_url)

        first_page = remove_notes(first_page.text)  # 去除网页里的注释
        if '文件取消' in first_page:
            return file_detail(CodeDetail.FILE_CANCELLED, pwd=pwd, url=share_url)

        # 这里获取下载直链 304 重定向前的链接
        if '输入密码' in first_page:  # 文件设置了提取码时
            if len(pwd) == 0:
                return file_detail(CodeDetail.LACK_PASSWORD, pwd=pwd, url=share_url)  # 没给提取码直接退出
            # data : 'action=downprocess&sign=AGZRbwEwU2IEDQU6BDRUaFc8DzxfMlRjCjTPlVkWzFSYFY7ATpWYw_c_c&p='+pwd,
            sign = re.search(r"sign=(\w+?)&", first_page).group(1)
            post_data = {'action': 'downprocess', 'sign': sign, 'p': pwd}
            link_info = self._post(self._host_url + '/ajaxm.php', post_data)  # 保存了重定向前的链接信息和文件名
            second_page = self._get(share_url)  # 再次请求文件分享页面，可以看见文件名，时间，大小等信息(第二页)
            if not link_info or not second_page.text:
                return file_detail(CodeDetail.NETWORK_ERROR, pwd=pwd, url=share_url)
            link_info = link_info.json()
            second_page = remove_notes(second_page.text)
            # 提取文件信息
            f_name = link_info['inf']
            f_size = re.search(r'大小.+?(\d[\d.]+\s?[BKM]?)<', second_page)
            f_size = f_size.group(1) if f_size else '0 M'
            f_time = re.search(r'class="n_file_infos">(.+?)</span>', second_page)
            f_time = time_format(f_time.group(1)) if f_time else time_format('0 小时前')
            f_desc = re.search(r'class="n_box_des">(.*?)</div>', second_page)
            f_desc = f_desc.group(1) if f_desc else ''
        else:  # 文件没有设置提取码时,文件信息都暴露在分享页面上
            para = re.search(r'<iframe.*?src="(.+?)"', first_page).group(1)  # 提取下载页面 URL 的参数
            # 文件名位置变化很多
            f_name = (re.search(r"<title>(.+?) - 蓝奏云</title>", first_page) or
                      re.search(r'<div class="filethetext".+?>([^<>]+?)</div>', first_page) or
                      re.search(r'<div style="font-size.+?>([^<>].+?)</div>', first_page) or
                      re.search(r"var filename = '(.+?)';", first_page) or
                      re.search(r'id="filenajax">(.+?)</div>', first_page) or
                      re.search(r'<div class="b"><span>([^<>]+?)</span></div>', first_page))
            f_name = f_name.group(1) if f_name else "未匹配到文件名"
            # 匹配文件时间，文件没有时间信息就视为今天，统一表示为 2020-01-01 格式
            f_time = re.search(r'>(\d+\s?[秒天分小][钟时]?前|[昨前]天\s?[\d:]+?|\d+\s?天前|\d{4}-\d\d-\d\d)<', first_page)
            f_time = time_format(f_time.group(1)) if f_time else time_format('0 小时前')
            # 匹配文件大小
            f_size = re.search(r'大小.+?(\d[\d.]+\s?[BKM]?)<', first_page)
            f_size = f_size.group(1) if f_size else '0 M'
            f_desc = re.search(r'文件描述.+?<br>\n?\s*(.*?)\s*</td>', first_page)
            f_desc = f_desc.group(1) if f_desc else ''
            first_page = self._get(self._host_url + para)
            if not first_page:
                return file_detail(CodeDetail.NETWORK_ERROR, name=f_name, time=f_time, size=f_size, desc=f_desc,
                                   pwd=pwd, url=share_url)
            first_page = remove_notes(first_page.text)
            # 一般情况 sign 的值就在 data 里，有时放在变量后面
            sign = re.search(r"'sign':(.+?),", first_page).group(1)
            if len(sign) < 20:  # 此时 sign 保存在变量里面, 变量名是 sign 匹配的字符
                sign = re.search(r"var {sign}\s*=\s*'(.+?)';".format(sign=sign), first_page).group(1)
            post_data = {'action': 'downprocess', 'sign': sign, 'ves': 1}
            link_info = self._post(self._host_url + '/ajaxm.php', post_data)
            if not link_info:
                return file_detail(CodeDetail.NETWORK_ERROR, name=f_name, time=f_time, size=f_size, desc=f_desc,
                                   pwd=pwd, url=share_url)
            else:
                link_info = link_info.json()
        # 这里开始获取文件直链
        if link_info['zt'] == 1:
            fake_url = link_info['dom'] + '/file/' + link_info['url']  # 假直连，存在流量异常检测
            download_page = self._get(fake_url, allow_redirects=False)
            if not download_page:
                return file_detail(CodeDetail.NETWORK_ERROR, name=f_name, time=f_time, size=f_size, desc=f_desc,
                                   pwd=pwd, url=share_url)
            download_page.encoding = 'utf-8'
            if '网络不正常' in download_page.text:  # 流量异常，要求输入验证码
                file_token = re.findall(r"'file':'(.+?)'", download_page.text)[0]
                direct_url = self._captcha_recognize(file_token)
                if not direct_url:
                    return file_detail(CodeDetail.CAPTCHA_ERROR, name=f_name, time=f_time, size=f_size, desc=f_desc,
                                       pwd=pwd, url=share_url)
            else:
                direct_url = download_page.headers['Location']  # 重定向后的真直链

            f_type = f_name.split('.')[-1]
            return file_detail(CodeDetail.SUCCESS,
                               name=f_name, size=f_size, type=f_type, time=f_time,
                               desc=f_desc, pwd=pwd, url=share_url, durl=direct_url)
        else:
            return file_detail(CodeDetail.FAILED, name=f_name, time=f_time, size=f_size, desc=f_desc, pwd=pwd,
                               url=share_url)

    def get_file_info_by_id(self, file_id) -> dict:
        """通过 id 获取文件信息"""
        info = self.get_share_info(file_id)
        if info['code'] != CodeDetail.SUCCESS:
            return file_detail(info['code'])
        return self.get_file_info_by_url(info['url'], info['pwd'])

    def get_durl_by_url(self, share_url, pwd='') -> dict:
        """通过分享链接获取下载直链"""
        file_info = self.get_file_info_by_url(share_url, pwd)
        if file_info['code'] != CodeDetail.SUCCESS:
            return direct_url_info(file_info['code'], '', '')
        return direct_url_info(CodeDetail.SUCCESS, file_info['name'], file_info['durl'])

    def get_durl_by_id(self, file_id) -> dict:
        """登录用户通过id获取直链"""
        info = self.get_share_info(file_id, is_file=True)  # 能获取直链，一定是文件
        return self.get_durl_by_url(info['url'], info['pwd'])

    def get_share_info(self, fid, is_file=True) -> dict:
        """获取文件(夹)提取码、分享链接"""
        post_data = {'task': 22, 'file_id': fid} if is_file else {'task': 18, 'folder_id': fid}  # 获取分享链接和密码用
        f_info = self._post(self._doupload_url, post_data)
        if not f_info:
            return ShareInfo(CodeDetail.NETWORK_ERROR)
        else:
            f_info = f_info.json()['info']

        # id 有效性校验
        if ('f_id' in f_info.keys() and f_info['f_id'] == 'i') or ('name' in f_info.keys() and not f_info['name']):
            return ShareInfo(CodeDetail.ID_ERROR)

        # onof=1 时，存在有效的提取码; onof=0 时不存在提取码，但是 pwd 字段还是有一个无效的随机密码
        pwd = f_info['pwd'] if f_info['onof'] == '1' else ''
        if 'f_id' in f_info.keys():  # 说明返回的是文件的信息
            url = f_info['is_newd'] + '/' + f_info['f_id']  # 文件的分享链接需要拼凑
            file_info = self._post(self._doupload_url, {'task': 12, 'file_id': fid})  # 文件信息
            if not file_info:
                return ShareInfo(CodeDetail.NETWORK_ERROR)
            name = file_info.json()['text']  # 无后缀的文件名(获得后缀又要发送请求,没有就没有吧,尽可能减少请求数量)
            desc = file_info.json()['info']
        else:
            url = f_info['new_url']  # 文件夹的分享链接可以直接拿到
            name = f_info['name']  # 文件夹名
            desc = f_info['des']  # 文件夹描述
        return ShareInfo(CodeDetail.SUCCESS, name=name, url=url, desc=desc, pwd=pwd)

    def set_passwd(self, fid, passwd='', is_file=True) -> int:
        """设置网盘文件(夹)的提取码"""
        # id 无效或者 id 类型不对应仍然返回成功 :(
        # 文件夹提取码长度 0-12 位  文件提取码 2-6 位
        passwd_status = 0 if passwd == '' else 1  # 是否开启密码
        if is_file:
            post_data = {"task": 23, "file_id": fid, "shows": passwd_status, "shownames": passwd}
        else:
            post_data = {"task": 16, "folder_id": fid, "shows": passwd_status, "shownames": passwd}
        result = self._post(self._doupload_url, post_data)
        if not result:
            return CodeDetail.NETWORK_ERROR
        return CodeDetail.SUCCESS if result.json()['zt'] == 1 else CodeDetail.FAILED

    def mkdir(self, parent_id, folder_name, desc='') -> int:
        """创建文件夹(同时设置描述)"""
        folder_name = folder_name.replace(' ', '_')  # 文件夹名称不能包含空格
        folder_name = name_format(folder_name)  # 去除非法字符
        folder_list = self.get_dir_list(parent_id)
        if find_by_name(folder_list, folder_name):  # 如果文件夹已经存在，直接返回 id
            return find_by_name(folder_list, folder_name)['id']
        raw_folders = self.get_move_folders()
        post_data = {"task": 2, "parent_id": parent_id or -1, "folder_name": folder_name,
                     "folder_description": desc}
        result = self._post(self._doupload_url, post_data)  # 创建文件夹
        if not result or result.json()['zt'] != 1:
            logger.debug("Mkdir {folder_name} error, {parent_id}".format(folder_name=folder_name, parent_id=parent_id))
            return CodeDetail.MKDIR_ERROR  # 正常时返回 id 也是 int，为了方便判断是否成功，网络异常或者创建失败都返回相同错误码
        # 允许再不同路径创建同名文件夹, 移动时可通过 get_move_paths() 区分
        for folder in self.get_move_folders():
            if not find_by_id(raw_folders, folder['id']):
                logger.debug(
                    "Mkdir {folder_name} #{id} in {parent_id}".format(folder_name=folder_name, parent_id=parent_id,
                                                                      id=folder['id']))
                return folder['id']
        logger.debug("Mkdir {folder_name} error, {parent_id}".format(folder_name=folder_name, parent_id=parent_id))
        return CodeDetail.MKDIR_ERROR

    def _set_dir_info(self, folder_id, folder_name, desc='') -> int:
        """重命名文件夹及其描述"""
        # 不能用于重命名文件，id 无效仍然返回成功
        folder_name = name_format(folder_name)
        post_data = {'task': 4, 'folder_id': folder_id, 'folder_name': folder_name, 'folder_description': desc}
        result = self._post(self._doupload_url, post_data)
        if not result:
            return CodeDetail.NETWORK_ERROR
        return CodeDetail.SUCCESS if result.json()['zt'] == 1 else CodeDetail.FAILED

    def rename_dir(self, folder_id, folder_name) -> int:
        """重命名文件夹"""
        # 重命名文件要开会员额
        info = self.get_share_info(folder_id, is_file=False)
        if info['code'] != CodeDetail.SUCCESS:
            return info['code']
        return self._set_dir_info(folder_id, folder_name, info['desc'])

    def set_desc(self, fid, desc, is_file=True) -> int:
        """设置文件(夹)描述"""
        if is_file:
            # 文件描述一旦设置了值，就不能再设置为空
            post_data = {'task': 11, 'file_id': fid, 'desc': desc}
            result = self._post(self._doupload_url, post_data)
            if not result:
                return CodeDetail.NETWORK_ERROR
            elif result.json()['zt'] != 1:
                return CodeDetail.FAILED
            return CodeDetail.SUCCESS
        else:
            # 文件夹描述可以置空
            info = self.get_share_info(fid, is_file=False)
            if info['code'] != CodeDetail.SUCCESS:
                return info['code']
            return self._set_dir_info(fid, info['name'], desc)

    def rename_file(self, file_id, filename):
        """允许会员重命名文件(无法修后缀名)"""
        post_data = {'task': 46, 'file_id': file_id, 'file_name': name_format(filename), 'type': 2}
        result = self._post(self._doupload_url, post_data)
        if not result:
            return CodeDetail.NETWORK_ERROR
        return CodeDetail.SUCCESS if result.json()['zt'] == 1 else CodeDetail.FAILED

    def get_move_folders(self) -> []:
        """获取全部文件夹 id-name 列表，用于移动文件至新的文件夹"""
        # 这里 file_id 可以为任意值,不会对结果产生影响
        result = [FolderId(name='LanZouCloud', id=-1)]
        resp = self._post(self._doupload_url, data={"task": 19, "file_id": -1})
        if not resp or resp.json()['zt'] != 1:  # 获取失败或者网络异常
            return result
        info = resp.json()['info'] or []  # 新注册用户无数据, info=None
        for folder in info:
            folder_id, folder_name = int(folder['folder_id']), folder['folder_name']
            result.append(FolderId(folder_name, folder_id))
        return result

    def get_move_paths(self) -> List[list]:
        """获取所有文件夹的绝对路径(耗时长)"""
        # 官方 bug, 可能会返回一些已经被删除的"幽灵文件夹"
        result = []
        root = [FolderId('LanZouCloud', -1)]
        result.append(root)
        resp = self._post(self._doupload_url, data={"task": 19, "file_id": -1})
        if not resp or resp.json()['zt'] != 1:  # 获取失败或者网络异常
            return result

        ex = ThreadPoolExecutor()  # 线程数 min(32, os.cpu_count() + 4)
        id_list = [int(folder['folder_id']) for folder in resp.json()['info']]
        task_list = [ex.submit(self.get_full_path, fid) for fid in id_list]
        for task in as_completed(task_list):
            result.append(task.result())
        return sorted(result)

    def move_file(self, file_id, folder_id=-1) -> int:
        """移动文件到指定文件夹"""
        # 移动回收站文件也返回成功(实际上行不通) (+_+)?
        post_data = {'task': 20, 'file_id': file_id, 'folder_id': folder_id}
        result = self._post(self._doupload_url, post_data)
        logger.debug("Move file {file_id} to {folder_id}".format(folder_id=folder_id, file_id=file_id))
        if not result:
            return CodeDetail.NETWORK_ERROR
        return CodeDetail.SUCCESS if result.json()['zt'] == 1 else CodeDetail.FAILED

    def move_folder(self, folder_id, parent_folder_id=-1) -> int:
        """移动文件夹(官方并没有直接支持此功能)"""
        if folder_id == parent_folder_id or parent_folder_id < -1:
            return CodeDetail.FAILED  # 禁止移动文件夹到自身，禁止移动到 -2 这样的文件夹(文件还在,但是从此不可见)

        folder = find_by_id(self.get_move_folders(), folder_id)
        if not folder:
            logger.debug("Not found folder :{folder_id}".format(folder_id=folder_id))
            return CodeDetail.FAILED

        if self.get_dir_list(folder_id):
            logger.debug("Found subdirectory in {folder}".format(folder=folder))
            return CodeDetail.FAILED  # 递归操作可能会产生大量请求,这里只移动单层文件夹

        info = self.get_share_info(folder_id, False)
        new_folder_id = self.mkdir(parent_folder_id, folder['name'], info['desc'])  # 在目标文件夹下创建同名文件夹

        if new_folder_id == CodeDetail.MKDIR_ERROR:
            return CodeDetail.FAILED
        elif new_folder_id == folder_id:  # 移动文件夹到同一目录
            return CodeDetail.FAILED

        self.set_passwd(new_folder_id, info['pwd'], False)  # 保持密码相同
        ex = ThreadPoolExecutor()
        task_list = [ex.submit(self.move_file, file['id'], new_folder_id) for file in self.get_file_list(folder_id)]
        for task in as_completed(task_list):
            if task.result() != CodeDetail.SUCCESS:
                return CodeDetail.FAILED
        self.delete(folder_id, False)  # 全部移动完成后删除原文件夹
        self.delete_rec(folder_id, False)
        return CodeDetail.SUCCESS

    def _upload_small_file(self, file_path, folder_id=-1, file_name=None) -> int:
        """绕过格式限制上传不超过 max_size 的文件"""
        if not os.path.isfile(file_path):
            return CodeDetail.PATH_ERROR

        need_delete = False  # 上传完成是否删除
        if not is_name_valid(os.path.basename(file_path)):  # 不允许上传的格式
            if self._limit_mode:  # 不允许绕过官方限制
                return CodeDetail.OFFICIAL_LIMITED
            file_path = let_me_upload(file_path)  # 添加了报尾的新文件
            need_delete = True

        # 文件已经存在同名文件就删除
        filename = file_name or name_format(os.path.basename(file_path))
        file_list = self.get_file_list(folder_id)
        if find_by_name(file_list, filename):
            self.delete(find_by_name(file_list, filename)['id'])
        logger.debug(f'Upload {file_path} to {folder_id}')

        file = open(file_path, 'rb')
        post_data = {
            "task": "1",
            "folder_id": str(folder_id),
            "id": "WU_FILE_0",
            "name": filename,
            "upload_file": (filename, file, 'application/octet-stream')
        }

        post_data = MultipartEncoder(post_data)
        tmp_header = self._headers.copy()
        tmp_header['Content-Type'] = post_data.content_type

        # MultipartEncoderMonitor 每上传 8129 bytes数据调用一次回调函数，问题根源是 httplib 库
        # issue : https://github.com/requests/toolbelt/issues/75
        # 上传完成后，回调函数会被错误的多调用一次(强迫症受不了)。因此，下面重新封装了回调函数，修改了接受的参数，并阻断了多余的一次调用
        self._upload_finished_flag = False  # 上传完成的标志

        progress = tqdm(unit='B', unit_scale=True, unit_divisor=1024, desc=filename, total=post_data.len)

        def _call_back(read_monitor):
            if read_monitor.len == read_monitor.bytes_read:
                self._upload_finished_flag = True
                progress.close()
            if not self._upload_finished_flag:
                progress.update(8192)

        monitor = MultipartEncoderMonitor(post_data, _call_back)

        result = self._post(up_url, data=monitor, headers=tmp_header, timeout=3600)
        if not result:  # 网络异常
            return CodeDetail.NETWORK_ERROR
        else:
            result = result.json()
        if result["zt"] != 1:
            logger.warning('Upload failed: {result}'.format(result=result))
            return CodeDetail.FAILED  # 上传失败

        if need_delete:
            file.close()
            os.remove(file_path)
        return CodeDetail.SUCCESS

    def _upload_big_file(self, file_path, dir_id, file_name=None):
        """上传大文件, 且使得回调函数只显示一个文件"""
        if self._limit_mode:  # 不允许绕过官方限制
            return CodeDetail.OFFICIAL_LIMITED

        file_size = os.path.getsize(file_path)  # 原始文件的字节大小
        file_name = file_name or os.path.basename(file_path)
        tmp_dir = os.path.dirname(file_path) + os.sep + '__' + '.'.join(file_name.split('.')[:-1])  # 临时文件保存路径
        record_file = tmp_dir + os.sep + file_name + '.record'  # 记录文件，大文件没有完全上传前保留，用于支持续传
        uploaded_size = 0  # 记录已上传字节数，用于回调函数

        if not os.path.exists(tmp_dir):
            os.makedirs(tmp_dir)
        if not os.path.exists(record_file):  # 初始化记录文件
            info = {'name': file_name, 'size': file_size, 'uploaded': 0, 'parts': []}
            with open(record_file, 'wb') as f:
                pickle.dump(info, f)
        else:
            with open(record_file, 'rb') as f:
                info = pickle.load(f)
                uploaded_size = info['uploaded']  # 读取已经上传的大小
                logger.debug("Find upload record: {uploaded_size}/{file_size}".format(uploaded_size=uploaded_size,
                                                                                      file_size=file_size))

        while uploaded_size < file_size:
            data_size, data_path = big_file_split(file_path, self._max_size, start_byte=uploaded_size)
            code = self._upload_small_file(data_path, dir_id)
            if code == CodeDetail.SUCCESS:
                uploaded_size += data_size  # 更新已上传的总字节大小
                info['uploaded'] = uploaded_size
                info['parts'].append(os.path.basename(data_path))  # 记录已上传的文件名
                with open(record_file, 'wb') as f:
                    logger.debug("Update record file: {uploaded_size}/{file_size}".format(uploaded_size=uploaded_size,
                                                                                          file_size=file_size))
                    pickle.dump(info, f)
            else:
                logger.debug("Upload data file failed: {data_path}".format(data_path=data_path))
                return CodeDetail.FAILED
            os.remove(data_path)  # 删除临时数据块
            min_s, max_s = self._upload_delay  # 设置两次上传间的延时，减小封号可能性
            sleep_time = uniform(min_s, max_s)
            logger.debug("Sleeping, Upload task will resume after {sleep_time}s...".format(sleep_time=sleep_time))
            sleep(sleep_time)

        # 全部数据块上传完成
        record_name = list(file_name.replace('.', ''))  # 记录文件名也打乱
        shuffle(record_name)
        record_name = name_format(''.join(record_name)) + '.txt'
        record_file_new = tmp_dir + os.sep + record_name
        os.rename(record_file, record_file_new)
        code = self._upload_small_file(record_file_new, dir_id)  # 上传记录文件
        if code != CodeDetail.SUCCESS:
            logger.debug("Upload record file failed: {record_file_new}".format(record_file_new=record_file_new))
            return CodeDetail.FAILED
        # 记录文件上传成功，删除临时文件
        shutil.rmtree(tmp_dir)
        logger.debug("Upload finished, Delete tmp folder:{tmp_dir}".format(tmp_dir=tmp_dir))
        return CodeDetail.SUCCESS

    def upload_file(self, file_path, folder_id=-1, file_name=None) -> int:
        """解除限制上传文件
        :param file_path:
        :param folder_id:
        """
        if not os.path.isfile(file_path):
            return CodeDetail.PATH_ERROR

        # 单个文件不超过 max_size 直接上传
        if os.path.getsize(file_path) <= self._max_size * 1048576:
            return self._upload_small_file(file_path, folder_id, file_name=file_name)

        # 上传超过 max_size 的文件
        if self._limit_mode:
            return CodeDetail.OFFICIAL_LIMITED

        folder_name = os.path.basename(file_path)  # 保存分段文件的文件夹名
        dir_id = self.mkdir(folder_id, folder_name, 'Big File')
        if dir_id == CodeDetail.MKDIR_ERROR:
            return CodeDetail.MKDIR_ERROR  # 创建文件夹失败就退出

        return self._upload_big_file(file_path, dir_id, file_name=file_name)

    def upload_dir(self, dir_path, folder_id=-1):
        """批量上传文件夹中的文件(不会递归上传子文件夹)
        :param folder_id: 网盘文件夹 id
        :param dir_path: 文件夹路径
        """
        if not os.path.isdir(dir_path):
            return CodeDetail.PATH_ERROR

        dir_name = dir_path.split(os.sep)[-1]
        dir_id = self.mkdir(folder_id, dir_name, '批量上传')
        if dir_id == CodeDetail.MKDIR_ERROR:
            return CodeDetail.MKDIR_ERROR

        for filename in os.listdir(dir_path):
            file_path = dir_path + os.sep + filename
            if not os.path.isfile(file_path):
                continue  # 跳过子文件夹
            code = self.upload_file(file_path, dir_id)
            if code != CodeDetail.SUCCESS:
                logger.error("upload error {}".format(dir_path))
        return CodeDetail.SUCCESS

    def down_file_by_url(self, share_url, pwd='', save_path='./download', file_name=None) -> int:
        """通过分享链接下载文件(需提取码)"""
        if not is_file_url(share_url):
            return CodeDetail.URL_INVALID
        if not os.path.exists(save_path):
            os.makedirs(save_path)

        info = self.get_durl_by_url(share_url, pwd)
        logger.debug('File direct url info: {info}'.format(info=info))
        if info['code'] != CodeDetail.SUCCESS:
            return info['code']

        resp = self._get(info['durl'], stream=True)
        if not resp:
            return CodeDetail.FAILED
        total_size = int(resp.headers['Content-Length'])

        file_path = save_path + os.sep + info['name']
        logger.debug('Save file to {file_path}'.format(file_path=file_path))
        if os.path.exists(file_path):
            now_size = os.path.getsize(file_path)  # 本地已经下载的文件大小
        else:
            now_size = 0
        chunk_size = 4096
        last_512_bytes = b''  # 用于识别文件是否携带真实文件名信息
        header = {**self._headers, 'Range': 'bytes=%d-' % now_size}
        resp = self._get(info['durl'], stream=True, headers=header)

        if resp is None:  # 网络异常
            return CodeDetail.FAILED
        if resp.status_code == 416:  # 已经下载完成
            return CodeDetail.SUCCESS

        progress = tqdm(unit='B', unit_scale=True, unit_divisor=1024, desc=info['name'], total=total_size)

        with open(file_path, "ab") as f:
            for chunk in resp.iter_content(chunk_size):
                if chunk:
                    f.write(chunk)
                    f.flush()
                    now_size += len(chunk)
                    if total_size - now_size < 512:
                        last_512_bytes += chunk
                    progress.update(len(chunk))
        progress.close()
        # 尝试解析文件报尾
        file_info = un_serialize(last_512_bytes[-512:])
        if file_info is not None and 'padding' in file_info:  # 大文件的记录文件也可以反序列化出 name,但是没有 padding
            real_name = file_name or file_info['name']
            new_file_path = save_path + os.sep + real_name
            logger.debug("Find meta info: {real_name}".format(real_name=real_name))
            if os.path.exists(new_file_path):
                os.remove(new_file_path)  # 存在同名文件则删除
            os.rename(file_path, new_file_path)
            with open(new_file_path, 'rb+') as f:
                f.seek(-512, 2)  # 截断最后 512 字节数据
                f.truncate()
        return CodeDetail.SUCCESS

    def down_file_by_id(self, fid, save_path='./Download', file_name=None) -> int:
        """登录用户通过id下载文件(无需提取码)"""
        info = self.get_share_info(fid, is_file=True)
        if info['code'] != CodeDetail.SUCCESS:
            return info['code']
        return self.down_file_by_url(info['url'], info['pwd'], save_path, file_name=file_name)

    def get_folder_info_by_url(self, share_url, dir_pwd='') -> dict:
        """获取文件夹里所有文件的信息"""
        if is_file_url(share_url):
            return FolderDetail(CodeDetail.URL_INVALID)
        try:
            html = requests.get(share_url, headers=self._headers).text
        except requests.RequestException:
            return FolderDetail(CodeDetail.NETWORK_ERROR)
        if '文件不存在' in html:
            return FolderDetail(CodeDetail.FILE_CANCELLED)
        if '请输入密码' in html and len(dir_pwd) == 0:
            return FolderDetail(CodeDetail.LACK_PASSWORD)
        try:
            # 获取文件需要的参数
            html = remove_notes(html)
            lx = re.findall(r"'lx':'?(\d)'?,", html)[0]
            t = re.findall(r"var [0-9a-z]{6} = '(\d{10})';", html)[0]
            k = re.findall(r"var [0-9a-z]{6} = '([0-9a-z]{15,})';", html)[0]
            # 文件夹的信息
            folder_id = re.findall(r"'fid':'?(\d+)'?,", html)[0]
            folder_name = re.findall(r"var.+?='(.+?)';\n.+document.title", html)[0]
            folder_time = re.findall(r'class="rets">([\d\-]+?)<a', html)[0]  # 日期不全 %m-%d
            folder_desc = re.findall(r'id="filename">(.+?)</span>', html)  # 无描述时无法完成匹配
            folder_desc = folder_desc[0] if len(folder_desc) == 1 else ''
        except IndexError:
            return FolderDetail(CodeDetail.FAILED)

        page = 1
        files = []
        while True:
            try:
                post_data = {'lx': lx, 'pg': page, 'k': k, 't': t, 'fid': folder_id, 'pwd': dir_pwd}
                resp = self._post(self._host_url + '/filemoreajax.php', data=post_data, headers=self._headers).json()
            except requests.RequestException:
                return FolderDetail(CodeDetail.NETWORK_ERROR)
            if resp['zt'] == 1:  # 成功获取一页文件信息
                for f in resp["text"]:
                    files.append(file_in_folder(
                        name=f["name_all"],  # 文件名
                        time=time_format(f["time"]),  # 上传时间
                        size=f["size"],  # 文件大小
                        type=f["name_all"].split('.')[-1],  # 文件格式
                        url=self._host_url + "/" + f["id"]  # 文件分享链接
                    ))
                page += 1  # 下一页
                continue
            elif resp['zt'] == 2:  # 已经拿到全部的文件信息
                break
            elif resp['zt'] == 3:  # 提取码错误
                return FolderDetail(CodeDetail.PASSWORD_ERROR)
            elif resp["zt"] == 4:
                continue
            else:
                return FolderDetail(CodeDetail.FAILED)  # 其它未知错误
        # 通过文件的时间信息补全文件夹的年份(如果有文件的话)
        if files:  # 最后一个文件上传时间最早，文件夹的创建年份与其相同
            folder_time = files[-1]['time'].split('-')[0] + '-' + folder_time
        else:  # 可恶，没有文件，日期就设置为今年吧
            folder_time = datetime.today().strftime('%Y-%m-%d')
        return FolderDetail(CodeDetail.SUCCESS,
                            file_in_folder(folder_name, folder_id, dir_pwd, folder_time, folder_desc, share_url),
                            files)

    def get_folder_info_by_id(self, folder_id):
        """通过 id 获取文件夹及内部文件信息"""
        info = self.get_share_info(folder_id, is_file=False)
        if info['code'] != CodeDetail.SUCCESS:
            return FolderDetail(info['code'])
        return self.get_folder_info_by_url(info['url'], info['pwd'])

    def _check_big_file(self, file_list):
        """检查文件列表,判断是否为大文件分段数据"""
        txt_files = find_filter(file_list, lambda f: f['name'].endswith('.txt') and 'M' not in f['size'])
        if txt_files and len(txt_files) == 1:  # 文件夹里有且仅有一个 txt, 很有可能是保存大文件的文件夹
            try:
                info = self.get_durl_by_url(txt_files[0]['url'])
            except AttributeError:
                info = self.get_durl_by_id(txt_files[0]['id'])
            if info['code'] != CodeDetail.SUCCESS:
                logger.debug("Big file checking: Failed")
                return None
            resp = self._get(info['durl'])
            info = un_serialize(resp.content) if resp else None
            if info is not None:  # 确认是大文件
                name, size, *_, parts = info.values()  # 真实文件名, 文件字节大小, (其它数据),分段数据文件名(有序)
                file_list = [find_by_name(file_list, p) for p in parts]
                if all(file_list):  # 分段数据完整
                    logger.debug("Big file checking: PASS , {name}, {size}".format(name=name, size=size))
                    return name, size, file_list
                logger.debug("Big file checking: Failed, Missing some data")
        logger.debug("Big file checking: Failed")
        return None

    def _down_big_file(self, name, total_size, file_list, save_path):
        """下载分段数据到一个文件，回调函数只显示一个文件
        支持大文件下载续传，下载完成后重复下载不会执行覆盖操作，直接返回状态码 SUCCESS
        """
        big_file = save_path + os.sep + name
        record_file = big_file + '.record'

        if not os.path.exists(save_path):
            os.makedirs(save_path)

        if not os.path.exists(record_file):  # 初始化记录文件
            info = {'last_ending': 0, 'finished': []}  # 记录上一个数据块结尾地址和已经下载的数据块
            with open(record_file, 'wb') as rf:
                pickle.dump(info, rf)
        else:  # 读取记录文件，下载续传
            with open(record_file, 'rb') as rf:
                info = pickle.load(rf)
                file_list = [f for f in file_list if f['name'] not in info['finished']]  # 排除已下载的数据块
                logger.debug("Find download record file: {info}".format(info=info))

        progress = tqdm(unit='B', unit_scale=True, unit_divisor=1024, desc=name, total=total_size)

        with open(big_file, 'ab') as bf:
            first_file = True
            for file in file_list:
                try:
                    durl_info = self.get_durl_by_url(file['url'])  # 分段文件无密码
                except AttributeError:
                    durl_info = self.get_durl_by_id(file['id'])
                if durl_info['code'] != CodeDetail.SUCCESS:
                    logger.debug("Can't get direct url: {file}".format(file=file))
                    return durl_info['code']
                # 准备向大文件写入数据
                file_size_now = os.path.getsize(big_file)
                down_start_byte = file_size_now - info['last_ending']  # 当前数据块上次下载中断的位置
                header = {**self._headers, 'Range': 'bytes=%d-' % down_start_byte}
                logger.debug("Download {name}, Range: {down_start_byte}-".format(name=file['name'],
                                                                                 down_start_byte=down_start_byte))
                resp = self._get(durl_info['durl'], stream=True, headers=header)

                if resp is None:  # 网络错误, 没有响应数据
                    return CodeDetail.FAILED
                if resp.status_code == 416:  # 下载完成后重复下载导致 Range 越界, 服务器返回 416
                    logger.debug("File {name} has already downloaded.".format(name=name))
                    os.remove(record_file)  # 删除记录文件
                    return CodeDetail.SUCCESS

                if first_file:
                    progress.update(file_size_now)
                first_file = False
                for chunk in resp.iter_content(4096):
                    if chunk:
                        file_size_now += len(chunk)
                        bf.write(chunk)
                        bf.flush()  # 确保缓冲区立即写入文件，否则下一次写入时获取的文件大小会有偏差
                        progress.update(len(chunk))

                # 一块数据写入完成，更新记录文件
                info['finished'].append(file['name'])
                info['last_ending'] = file_size_now
                with open(record_file, 'wb') as rf:
                    pickle.dump(info, rf)
                logger.debug("Update download record info: {info}".format(info=info))
            # 全部数据块下载完成, 记录文件可以删除
            logger.debug("Delete download record file: {record_file}".format(record_file=record_file))
            os.remove(record_file)
            progress.close()
        return CodeDetail.SUCCESS

    def down_dir_by_url(self, share_url, dir_pwd='', save_path='./download', *, mkdir=True) -> int:
        """通过分享链接下载文件夹
        :param dir_pwd:
        :param share_url:
        :param save_path 文件夹保存路径
        :param mkdir 是否在 save_path 下创建与远程文件夹同名的文件夹
        """
        folder_detail = self.get_folder_info_by_url(share_url, dir_pwd)
        if folder_detail['code'] != CodeDetail.SUCCESS:  # 获取文件信息失败
            return folder_detail['code']

        # 检查是否大文件分段数据
        info = self._check_big_file(folder_detail['files'])
        if info is not None:
            return self._down_big_file(*info, save_path)

        if mkdir:  # 自动创建子文件夹
            save_path = save_path + os.sep + folder_detail['folder']['name']
            if not os.path.exists(save_path):
                os.makedirs(save_path)

        # 不是大文件分段数据,直接下载
        for file in folder_detail['files']:
            code = self.down_file_by_url(file['url'], dir_pwd, save_path)
            logger.debug('Download file result: Code:{code}, File: {file}'.format(code=code, file=file))
            if code != CodeDetail.SUCCESS:
                logger.error("download error {}".format(share_url))

        return CodeDetail.SUCCESS

    def down_dir_by_id(self, folder_id, save_path='./download', *, mkdir=True) -> int:
        """登录用户通过id下载文件夹"""
        file_list = self.get_file_list(folder_id)
        if len(file_list) == 0:
            return CodeDetail.FAILED

        # 检查是否大文件分段数据
        info = self._check_big_file(file_list)
        if info is not None:
            return self._down_big_file(*info, save_path)

        if mkdir:  # 自动创建子目录
            share_info = self.get_share_info(folder_id, False)
            if share_info['code'] != CodeDetail.SUCCESS:
                return share_info['code']
            save_path = save_path + os.sep + share_info['name']
            if not os.path.exists(save_path):
                logger.debug("Mkdir {save_path}".format(save_path=save_path))
                os.makedirs(save_path)

        for file in file_list:
            code = self.down_file_by_id(file['id'], save_path)
            logger.debug('Download file result: Code:{code}, File: {file}'.format(code=code, file=file))
            if code != CodeDetail.SUCCESS:
                logger.error("download error folder_id={}".format(folder_id))

        return CodeDetail.SUCCESS

    def down_by_url(self, share_url, dir_pwd='', save_path='./download', *, mkdir=True) -> int:
        if is_file_url(share_url):
            return self.down_file_by_url(share_url, pwd=dir_pwd, save_path=save_path)
        elif is_folder_url(share_url):
            return self.down_dir_by_url(share_url, dir_pwd=dir_pwd, save_path=save_path, mkdir=mkdir)

    def sync_files(self,
                   path_root,
                   folder_id,
                   only_directory=False,
                   overwrite=False,
                   filter_fun=None,
                   remove_local=False
                   ):
        """
        将本地的文件同步到云端，单向同步
        :param path_root: 本地路径
        :param folder_id: 云端路径
        :param only_directory: 是否只同步文件夹
        :param overwrite: 是否需要覆盖重写
        :param filter_fun: 针对部分文件需要过滤
        :param remove_local: 同步完成后是否删除本地文件
        :return: 文件到folder_id的映射关系
        """
        yun_dir_list = self.get_dir_list(folder_id)
        yun_file_list = self.get_file_list(folder_id)
        yun_dir_dict = dict([(yun['name'], yun['id']) for yun in yun_dir_list])
        yun_file_dict = dict([(yun['name'], yun['id']) for yun in yun_file_list])

        file_dict = {}
        for file in (os.listdir(path_root)):
            local_path = os.path.join(path_root, file)
            # 根据传入的函数进行过滤，某些文件可以不同步
            if filter_fun is not None and filter_fun(local_path):
                continue

            # 文件夹同步，支持递归同步
            if os.path.isdir(local_path):
                if file in yun_dir_dict.keys():
                    yun_id = yun_dir_dict[file]
                else:
                    yun_id = self.mkdir(parent_id=folder_id, folder_name=file, desc=file)
                file_dict[local_path] = yun_id
                file_dict.update(self.sync_directory(local_path, yun_id))
            else:
                # 只同步文件夹
                if only_directory:
                    continue

                # 文件在云端已存在，如果覆盖重写，删除云端文件，重新上传
                if file in yun_file_dict.keys():
                    if overwrite:
                        self.delete(yun_file_dict[file], is_file=True)
                        yun_id = self.upload_file(file_path=local_path, folder_id=folder_id)
                    else:
                        yun_id = yun_file_dict[file]
                else:
                    yun_id = self.upload_file(file_path=local_path, folder_id=folder_id)
                file_dict[local_path] = yun_id
                if remove_local:
                    os.remove(local_path)

        return file_dict

    def sync_directory(self, path_root, folder_id):
        return self.sync_files(path_root, folder_id, only_directory=True)


def download(url, dir_pwd='./download'):
    downer = LanZouCloud()
    downer.ignore_limits()
    downer.down_by_url(url, save_path=dir_pwd)



import os


def wget(url, output=None, append=None, debug=None, quiet=None, verbose=None, noverbose=None):
    kwargs = {
        "output-file": output,
        "append-output": append,
        "debug": debug,
        "quiet": quiet,
        "verbose": verbose,
        "no-verbose": noverbose
    }

    args = ['--{}={}'.format(key, value) for key, value in kwargs.items() if value is not None]

    command = "wget {args} {url}".format(url=url, args=' '.join(args))
    res = os.system(command)
    return res



import logging
import os.path

from notetool.secret import read_secret

from github import Github


def all_file_to_github(git_path, repo_str=None, repo=None, access_tokens=None, recursive=False):
    if repo is None:
        access_tokens = access_tokens or read_secret(cate1='github', cate2='access_tokens', cate3='pygithub')
        g = Github(access_tokens)
        repo = g.get_repo(repo_str)

    all_files = []
    try:
        contents = repo.get_contents(git_path)
    except Exception as e:
        contents = repo.get_contents("")
        logging.warning(f"{git_path} not exists:{e}")
    while contents:
        file_content = contents.pop(0)
        if file_content.type == "dir":
            if recursive:
                contents.extend(repo.get_contents(file_content.path))
        else:
            all_files.append(file_content.path)
    return all_files


def upload_data_to_github(content, git_path, repo_str):
    access_tokens = read_secret(cate1='github', cate2='access_tokens', cate3='pygithub')
    g = Github(access_tokens)
    repo = g.get_repo(repo_str)
    all_files = all_file_to_github(os.path.dirname(git_path), repo=repo)

    if git_path in all_files:
        contents = repo.get_contents(git_path)
        repo.update_file(contents.path, "committing files", content, contents.sha, branch="master")
        logging.info(f'{git_path} UPDATED')
    else:
        repo.create_file(git_path, "committing files", content, branch="master")
        logging.info(f'{git_path} CREATED')


def upload_file_to_github(file_path, *args, **kwargs):
    with open(file_path, 'r') as file:
        content = file.read()
        upload_data_to_github(content=content, *args, **kwargs)



import argparse
import os

from notedrive.aria2 import Aria2JsonRpc

ID = 'a2jrg'


def parse_args(args=None):
    """
    Parse command-line arguments.
    """
    parser = argparse.ArgumentParser(description='Queue downloads via the Aria2 RPC interface.')
    parser.add_argument('uris', metavar='<URI>', nargs='*', help='File URIs. These must all point to the same file.')

    group = parser.add_argument_group('server options')
    Aria2JsonRpc.add_server_arguments(group)

    group = parser.add_argument_group('download options')
    group.add_argument('-o', '--options', nargs=argparse.REMAINDER,
                       help='Aria2 options as key-value pairs separated by "=".')

    group = parser.add_argument_group('commands')
    group.add_argument('--pause-all', action='store_true', help='Pause all downloads.')
    group.add_argument('--unpause-all', action='store_true', help='Unpause all downloads.')
    group.add_argument('--purge', action='store_true', help='Purge download results.')

    group = parser.add_argument_group('other options')
    group.add_argument('-w', '--wait', action='store_true', help='Wait for the download to finish')
    group.add_argument('-i', '--interval', type=int, default=1,
                       help='Status query interval while waiting, in seconds. Default: %(default)s.')
    group.add_argument('--status', action='store_true', help='Print global status report.')
    group.add_argument('--version', action='store_true', help='Print server version.')

    return parser.parse_args(args)


# --------------------------------- Actions ---------------------------------- #

def queue_uris(a2jr, args):
    """
    Queue URIs given an Aria2JsonRpc object and arguments. This will set the
    current directory to the download destination if one has not yet been given.
    """
    options = dict()
    if args.options:
        for option in args.options:
            key, value = option.split('=', 1)
            options[key] = value
    if 'dir' not in options:
        options['dir'] = os.getcwd()
    if args.wait:
        a2jr.queue_uris(args.uris, options, interval=args.interval)
    else:
        a2jr.queue_uris(args.uris, options)


# ----------------------------------- Main ----------------------------------- #

def main(args=None):
    """
    Parse command-line arguments and send the command to the RPC server.
    """
    pargs = parse_args(args)

    a2jr = Aria2JsonRpc.a2jr_from_args(ID, pargs)

    if pargs.pause_all:
        a2jr.pauseAll()

    elif pargs.unpause_all:
        a2jr.unpauseAll()

    if pargs.status:
        a2jr.print_global_status()

    if pargs.version:
        version = a2jr.getVersion()
        print('version: {}'.format(version['version']))
        print('enabled features:')
        for feature in version['enabledFeatures']:
            print(' ', feature)

    if pargs.uris:
        queue_uris(a2jr, pargs)

    if pargs.purge:
        a2jr.purgeDownloadResult()


if __name__ == '__main__':
    main()



import heapq
import struct
from ipaddress import IPv4Address
from operator import itemgetter


class Node:
    """
    Simple object to encapsulate the concept of a Node (minimally an ID, but
    also possibly an IP and port if this represents a node on the network).
    This class should generally not be instantiated directly, as it is a low
    level construct mostly used by the router.
    """

    def __init__(self, node_id: bytes, ip: str = None, port: int = None):
        """
        Create a Node instance.

        Args:
            node_id (bytes): A value between 0 and 2^160
            ip (string): Optional IP address where this Node lives
            port (int): Optional port for this Node (set when IP is set)
        """
        self.id = node_id
        self.ip = ip
        self.port = port
        self.long_id = int(node_id.hex(), 16)

    def same_home_as(self, node):
        return self.ip == node.ip and self.port == node.port

    def distance_to(self, node):
        """
        Get the distance between this node and another.
        """
        return self.long_id ^ node.long_id

    def __iter__(self):
        """
        Enables use of Node as a tuple - i.e., tuple(node) works.
        """
        return iter([self.id, self.ip, self.port])

    def __repr__(self):
        return repr([self.long_id, self.ip, self.port])

    def __str__(self):
        return "%s:%s" % (self.ip, str(self.port))

    @property
    def packed_ip_port(self):
        return IPv4Address(self.ip).packed + struct.pack("!H", self.port)

    @property
    def packed(self):
        return self.id + self.packed_ip_port


class NodeHeap:
    """
    A heap of nodes ordered by distance to a given node.
    """

    def __init__(self, node, maxsize):
        """
        Constructor.

        @param node: The node to measure all distnaces from.
        @param maxsize: The maximum size that this heap can grow to.
        """
        self.node = node
        self.heap = []
        self.contacted = set()
        self.maxsize = maxsize

    def remove(self, peers):
        """
        Remove a list of peer ids from this heap.  Note that while this
        heap retains a constant visible size (based on the iterator), it's
        actual size may be quite a bit larger than what's exposed.  Therefore,
        removal of nodes may not change the visible size as previously added
        nodes suddenly become visible.
        """
        peers = set(peers)
        if not peers:
            return
        nheap = []
        for distance, node in self.heap:
            if node.id not in peers:
                heapq.heappush(nheap, (distance, node))
        self.heap = nheap

    def get_node(self, node_id):
        for _, node in self.heap:
            if node.id == node_id:
                return node
        return None

    def have_contacted_all(self):
        return len(self.get_uncontacted()) == 0

    def get_ids(self):
        return [n.id for n in self]

    def mark_contacted(self, node):
        self.contacted.add(node.id)

    def popleft(self):
        return heapq.heappop(self.heap)[1] if self else None

    def push(self, nodes):
        """
        Push nodes onto heap.

        @param nodes: This can be a single item or a C{list}.
        """
        if not isinstance(nodes, list):
            nodes = [nodes]

        for node in nodes:
            if node not in self:
                distance = self.node.distance_to(node)
                heapq.heappush(self.heap, (distance, node))

    def __len__(self):
        return min(len(self.heap), self.maxsize)

    def __iter__(self):
        nodes = heapq.nsmallest(self.maxsize, self.heap)
        return iter(map(itemgetter(1), nodes))

    def __contains__(self, node):
        for _, other in self.heap:
            if node.id == other.id:
                return True
        return False

    def get_uncontacted(self):
        return [n for n in self if n.id not in self.contacted]



import os

from notesecret import read_secret
from tqdm import tqdm

from lanzou.api import LanZouCloud
from lanzou.api.models import FolderList
from lanzou.api.types import *


class ProgressWrap:
    def __init__(self, callback: tqdm = None):
        self.callback: tqdm = callback
        self.last_size = 0

    def init(self, file_name, total_size):
        if self.callback is None:
            self.callback = tqdm(unit='B', unit_scale=True, unit_divisor=1024, desc=file_name, total=total_size)

    def update(self, now_size):
        self.callback.update(now_size - self.last_size)
        self.last_size = now_size


class LanZouDrive(LanZouCloud):
    def __init__(self):
        super(LanZouDrive, self).__init__()

    def _upload_small_file(self, *args, callback=None, **kwargs) -> int:
        wrap = ProgressWrap()

        def clb(file_name, total_size, now_size):
            wrap.init(file_name, total_size)
            wrap.update(now_size)

        return super(LanZouDrive, self)._upload_small_file(*args, callback=clb, **kwargs)

    def down_file_by_url(self, *args, callback=None, **kwargs) -> int:
        """通过分享链接下载文件(需提取码)
        :param callback 用于显示下载进度 callback(file_name, total_size, now_size)
        :param overwrite 文件已存在时是否强制覆盖
        :param downloaded_handler 下载完成后进一步处理文件的回调函数 downloaded_handle(file_path)
        """

        wrap = ProgressWrap()

        def clb(file_name, total_size, now_size):
            wrap.init(file_name, total_size)
            wrap.update(now_size)

        return super(LanZouDrive, self).down_file_by_url(*args, callback=clb, **kwargs)

    def login_by_cookie(self, cookie: dict = None, ylogin=None, phpdisk_info=None) -> int:
        """通过cookie登录"""
        cookie = cookie or {
            'ylogin': read_secret("drive", "lanzou", 'ylogin', value=ylogin),
            'phpdisk_info': read_secret("drive", "lanzou", 'phpdisk_info', value=phpdisk_info)
        }
        return super(LanZouDrive, self).login_by_cookie(cookie)

    def sync_files(self, path_root, folder_id, only_directory=False, overwrite=False, filter_fun=None,
                   remove_local=False):
        """
        将本地的文件同步到云端，单向同步
        :param path_root: 本地路径
        :param folder_id: 云端路径
        :param only_directory: 是否只同步文件夹
        :param overwrite: 是否需要覆盖重写
        :param filter_fun: 针对部分文件需要过滤
        :param remove_local: 同步完成后是否删除本地文件
        :return: 文件到folder_id的映射关系
        """
        yun_dir_list = self.get_dir_list(folder_id)
        yun_file_list = self.get_file_list(folder_id)
        yun_dir_dict = dict([(yun.name, yun.id) for yun in yun_dir_list])
        yun_file_dict = dict([(yun.name, yun.id) for yun in yun_file_list])

        file_dict = {}
        for file in (os.listdir(path_root)):
            local_path = os.path.join(path_root, file)
            # 根据传入的函数进行过滤，某些文件可以不同步
            if filter_fun is not None and filter_fun(local_path):
                continue

            # 文件夹同步，支持递归同步
            if os.path.isdir(local_path):
                if file in yun_dir_dict.keys():
                    yun_id = yun_dir_dict[file]
                else:
                    yun_id = self.mkdir(parent_id=folder_id, folder_name=file, desc=file)
                file_dict[local_path] = yun_id
                file_dict.update(self.sync_files(local_path, yun_id, only_directory=only_directory,
                                                 overwrite=overwrite, filter_fun=filter_fun, remove_local=remove_local))
            else:
                # 只同步文件夹
                if only_directory:
                    continue
                # 文件在云端已存在，如果覆盖重写，删除云端文件，重新上传
                if file in yun_file_dict.keys():
                    if overwrite:
                        self.delete(yun_file_dict[file], is_file=True)
                        yun_id = self.upload_file(file_path=local_path, folder_id=folder_id)
                    else:
                        yun_id = yun_file_dict[file]
                    if remove_local:
                        os.remove(local_path)
                else:
                    yun_id = self.upload_file(file_path=local_path, folder_id=folder_id)
                file_dict[local_path] = yun_id

        return file_dict

    def sync_directory(self, path_root, folder_id, *args, **kwargs):
        return self.sync_files(path_root, folder_id, *args, **kwargs)

    def get_full_path(self, folder_id=-1) -> FolderList:
        """获取文件夹完整路径"""
        path_list = FolderList()
        path_list.append(FolderId('LanZouCloud', -1))
        post_data = {'task': 47, 'folder_id': folder_id}
        resp = self._post(self._doupload_url, post_data)
        if not resp:
            return path_list
        for folder in resp.json()['info']:
            if folder['folderid'] and folder['name']:  # 有时会返回无效数据, 这两个字段中某个为 None
                path_list.append(FolderId(id=int(folder['folderid']), name=folder['name']))
        return path_list


def download(url, dir_pwd='./download', pwd=''):
    downer = LanZouDrive()
    downer.ignore_limits()
    downer.down_file_by_url(url, save_path=dir_pwd, pwd=pwd)



import sys
from os import path

from notebuild.tool import read_version
from setuptools import find_packages, setup

version_path = path.join(path.abspath(path.dirname(__file__)), 'script/__version__.md')

version = read_version(version_path)


install_requires = ['requests', 'notesecret', 'demjson', 'tqdm', 'requests_toolbelt', 'lanzou-api']


setup(name='notedrive',
      version=version,
      description='notedrive',
      author='niuliangtao',
      author_email='1007530194@qq.com',
      url='https://github.com/1007530194',
      packages=find_packages(),
      install_requires=install_requires,
      )
